from .anonymize import Anonymize
from .secrets import Secrets
from .prompt_injections import PromptInjections
from .stop_input_substrings import StopInputSubstrings
from .dos_tokens import DoSTokens
from .malware_url import MalwareInputURL
from .toxicity import ToxicityInput
from .harmful import HarmfulInput
from .text_quality import TextQualityInput
from .coding_language import CodingLanguageInput
from .regex import RegexInput

__all__ = [
    "Anonymize",
    "Secrets",
    "PromptInjections",
    "DoSTokens",
    "StopInputSubstrings",
    "MalwareInputURL",
    "ToxicityInput",
    "HarmfulInput",
    "TextQualityInput",
    "CodingLanguageInput",
    "RegexInput"
]


import re
import json
import csv

from typing import List, Union, Dict, Any

from transformers import PreTrainedModel, PreTrainedTokenizer, BitsAndBytesConfig
import torch
from transformers import AutoModelForCausalLM, AutoTokenizer, pipeline

from jsonformer.format import highlight_values
from jsonformer.main import Jsonformer

from unstructured.partition.auto import partition
from unstructured.documents.elements import NarrativeText
from unstructured.partition.text_type import sentence_count


class DatasetGenerator:
    def __init__(
        self,
        file_path: str,
        output_path: str,
        *,
        model="OpenAssistant/falcon-7b-sft-mix-2000",
        tokenizer="OpenAssistant/falcon-7b-sft-mix-2000b",
        load_in_4bit=True,
        debug: bool = False,
        max_array_length: int = 256,
        max_number_tokens: int = 64,
        temperature: float = 0.3,
        max_string_token_length: int = 1024,
    ):
        if load_in_4bit:
            print("Loading in 4bit...")
            bnb_config = BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_quant_type="nf4",
                bnb_4bit_compute_dtype=torch.float16,
            )
            self.model = AutoModelForCausalLM.from_pretrained(
                model, quantization_config=bnb_config, trust_remote_code=True
            )
            self.tokenizer = AutoTokenizer.from_pretrained(model)
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(
                model, padding_side="left", use_fast=True, max_length=1024, use_cache=True
            )
            self.model = AutoModelForCausalLM.from_pretrained(
                tokenizer, device_map="auto", torch_dtype=torch.bfloat16, use_cache=True
            )
        self.max_array_length = max_array_length
        self.max_number_tokens = max_number_tokens
        self.temperature = temperature
        self.max_string_token_length = max_string_token_length
        self.debug = debug

        self.command = "You are an API that converts bodies of text into five highly contextual questions and answer pairs into a JSON format from the following text: "
        self.schema = " Generate five question and answer pairs based on the following schema:"
        self.json_schema = {
            "type": "object",
            "properties": {
                "qa_pair1": {
                    "type": "object",
                    "properties": {
                        "question": {"type": "string"},
                        "answer": {"type": "string"},
                    },
                },
                "qa_pair2": {
                    "type": "object",
                    "properties": {
                        "question": {"type": "string"},
                        "answer": {"type": "string"},
                    },
                },
                "qa_pair3": {
                    "type": "object",
                    "properties": {
                        "question": {"type": "string"},
                        "answer": {"type": "string"},
                    },
                },
            },
        }
        self.file_path = file_path
        self.output_path = output_path

    def generate_dataset(self):
        results = self.partition_file(self.file_path)
        qa_pairs = self.generate_pairs(results)
        json_output = self.convert_json(qa_pairs)
        self.validate_json(json_output)
        self.save_json(json_output, self.output_path)

    def partition_file(self, file_path):
        results = []
        if file_path:
            elements = partition(filename=file_path, strategy="hi_res")
            print("Elements", len(elements))
            for element in elements:
                if isinstance(element, NarrativeText) and sentence_count(element.text) > 1:
                    results.append(element.text)
        return results

    def generate_pairs(self, results):
        qa_pairs_arr = []
        print(len(results))
        if len(results) > 0:
            for text in results:
                prompt = self.generate_prompt(text)
                builder = Jsonformer(
                    model=self.model,
                    tokenizer=self.tokenizer,
                    json_schema=self.json_schema,
                    prompt=prompt,
                    max_string_token_length=self.max_string_token_length,
                    max_array_length=self.max_array_length,
                    max_number_tokens=self.max_number_tokens,
                    temperature=self.temperature,
                )
                output = builder()
                qa_pairs_arr.append(output)
                print(output)
                print("Size of array: ", len(qa_pairs_arr))
        return qa_pairs_arr

    def generate_prompt(self, input):
        prompt_intro = "Below is an instruction that describes a task. Write a response that appropriately completes the request."
        prompt_instruction = """

        ### Instruction:
        Heed the following rules:
        - Generate a highly contextual question and answer pairs from the following context
        - Avoid leading questions, or questions with the answer explicitly in them
        - Each question must be unique with no duplicates
        - For each answer, only output answers that can be explicitly referenced in the following context.

        You are an API that converts bodies of text into a unique question and answer pairs into a JSON format
        from the following text, while sticking to the aforementioned rules:
        """
        INPUT_KEY = "Context:"
        RESPONSE_KEY = "category=closed_qa"

        prompt_schema = "Generate a different question and answer pair sticking to the aforementioned rules and based on the following schema"
        prompt_full = """{intro}
        {instruction}
        {response_key}

        {input_key}
        {input}

        {schema}
        """.format(
            intro=prompt_intro,
            instruction=prompt_instruction,
            input_key=INPUT_KEY,
            input=input,
            response_key=RESPONSE_KEY,
            schema=prompt_schema,
        )
        return prompt_full

    def convert_json(self, json_data):
        """Converts the given JSON data into a single JSON output with just the question and answer pairs as objects.

        Args:
            json_data: The JSON data to be converted.

        Returns:
            A JSON object with the question and answer pairs as objects.
        """
        qa_pairs = []
        for item in json_data:
            for qa_pair in item.keys():
                question = item[qa_pair]["question"]
                answer = item[qa_pair]["answer"]

                # Clean up the question and answer pairs for any special characters.
                question = re.sub(r"[^\w\s,.!?]", "", question)
                answer = re.sub(r"[^\w\s,.!?]", "", answer)

                qa_pairs.append({"question": question, "answer": answer})
        return json.dumps(qa_pairs, indent=4)

    def validate_json(self, json_data):
        """Validates that the given JSON data is valid JSON.

        Args:
            json_data: The JSON data to be validated.

        Raises:
            Exception: If the given JSON data is not valid JSON.
        """
        try:
            json.loads(json_data)
        except Exception:
            raise Exception("The given JSON data is not valid JSON.")

    def save_json(self, json_data, filename):
        """Saves the given JSON data to a file.

        Args:
            json_data: The JSON data to be saved.
            filename: The filename to save the JSON data to.
        """
        # Writing to sample.json
        with open(filename, "w") as outfile:
            outfile.write(json_data)



from guardrail.metrics.offline_evals.bias import Bias
from guardrail.metrics.offline_evals.textstat import Textstat
from guardrail.metrics.offline_evals.toxicity import Toxicity
from guardrail.metrics.offline_evals.relevance import Relevance
from guardrail.metrics.offline_evals.injections import PromptInjection
from guardrail.metrics.offline_evals.sentiment import Sentiment



import logging
import re

logger = logging.getLogger(__name__)

class TextUtilities:
    def __init__(self):
        self.pat_1_10 = re.compile(r"\s*([1-9][0-9]*)\s*")

    def re_1_10_score(self, str_val):
        matches = self.pat_1_10.fullmatch(str_val)
        if not matches:
            matches = re.search('[1-9][0-9]*', str_val)
            if not matches:
                logger.warning(f"1-10 score regex failed to match on: '{str_val}'")
                return -10  # so this will be reported as -1 after division by 10

        return int(matches.group())


import json
import os
import numpy as np
import jsonlines
import pdb
import asyncio

from guardrail.firewall.output_detectors.factuality_detector.utils import OpenAIEmbedWrapper

class local_search():
    def __init__(self, snippet_cnt, data_link, data=None, embedding_link=None):
        self.snippet_cnt = snippet_cnt
        self.data_link = data_link
        self.embedding_link = embedding_link
        self.openai_embed = OpenAIEmbedWrapper()
        self.data = None
        self.embedding = None
        asyncio.run(self.init_async())
    
    async def init_async(self):
        print("init local search")
        self.load_data_by_link()
        if self.embedding_link is None:
            await self.calculate_embedding()
        else:
            self.load_embedding_by_link()
        print("loaded data and embedding")

    def add_suffix_to_json_filename(self, filename):
        base_name, extension = os.path.splitext(filename)
        return base_name + '_embed' + extension

    def load_data_by_link(self):
        #load data from json link
        self.data = []
        #self.data = json.load(open(self.data_link, 'r'))
        with jsonlines.open(self.data_link) as reader:
            for obj in reader:
                self.data.append(obj['text'])

    def load_embedding_by_link(self):
        self.embedding = []
        #self.embedding = json.load(open(self.embedding_link, 'r'))
        with jsonlines.open(self.embedding_link) as reader:
            for obj in reader:
                self.embedding.append(obj)
    
    def save_embeddings(self):
        #json.dump(self.embedding, open(self.add_suffix_to_json_filename(self.data_link), 'w'))
        with jsonlines.open(self.add_suffix_to_json_filename(self.data_link), mode='w') as writer:
            writer.write_all(self.embedding)

    async def calculate_embedding(self):
        result = await self.openai_embed.process_batch(self.data,retry=3)
        self.embedding = [emb["data"][0]["embedding"] for emb in result]
        self.save_embeddings()

    async def search(self, query):
        result = await self.openai_embed.create_embedding(query)
        query_embed = result["data"][0]["embedding"]
        dot_product = np.dot(self.embedding, query_embed)
        sorted_indices = np.argsort(dot_product)[::-1]
        top_k_indices = sorted_indices[:self.snippet_cnt]
        return [{"content":self.data[i],"source":"local"} for i in top_k_indices]

    async def run(self, queries):
        flattened_queries = []
        for sublist in queries:
            if sublist is None:
                sublist = ['None', 'None']
            for item in sublist:
                flattened_queries.append(item)
        
        snippets = await asyncio.gather(*[self.search(query) for query in flattened_queries])
        snippets_split = [snippets[i] + snippets[i+1] for i in range(0, len(snippets), 2)]
        return snippets_split


import logging
import json 
from abc import abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import torch
from colorama import Fore

from pydantic import Extra, Field, BaseModel, PrivateAttr
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from ctransformers import AutoModelForCausalLM

from guardrail.guardchain.agent.message import BaseMessage, AIMessage
from guardrail.guardchain.tools.base_tool import Tool

from guardrail.guardchain.utils import print_with_color

from guardrail.guardchain.models.jsonformer import Jsonformer

logger = logging.getLogger(__name__)


class Generation(BaseModel):
    """Output of a single generation."""

    message: BaseMessage
    """Generated text output."""

    generation_info: Optional[Dict[str, Any]] = None
    """Raw generation info response from the provider."""
    """May include things like reason for finishing (e.g., in OpenAI)."""
    # TODO: add log probs


class LLMResult(BaseModel):
    """Class that contains all relevant information for an LLM Result."""

    generations: List[Generation]
    """List of the things generated. This is List[List[]] because
    each input could have multiple generations."""
    llm_output: Optional[dict] = None
    """For arbitrary LLM provider specific output."""


class EmbeddingResult(BaseModel):
    texts: List[str]
    embeddings: List[List[float]]

class GGMLHuggingFaceModel():
    """Wrapper around HuggingFace's Transformers language models.

    To use, you should have the `transformers` library installed, and the
    appropriate model and tokenizer should be loaded.

    Any parameters that are valid to be passed to the `generate` call can be passed
    in, even if not explicitly saved on this class.
    """

    model_name: str = "TheBloke/WizardLM-13B-V1.2-GGML"
    """Model name to use."""
    model_file: str = "wizardlm-13b-v1.2.ggmlv3.q4_1.bin"
    """GGML Model Bin File"""
    gpu_layers: int = 50
    temperature: float = 0.7
    """What sampling temperature to use."""
    model_kwargs: Dict[str, Any] = None
    """Holds any model parameters valid for `generate` call not explicitly specified."""
    request_timeout: Optional[Union[float, Tuple[float, float]]] = None
    """Timeout for requests to completion API. Default is 600 seconds."""
    max_retries: int = 6
    """Maximum number of retries to make when generating."""
    n: int = 1
    """Number of completions to generate for each prompt."""
    max_tokens: Optional[int] = None
    """Maximum number of tokens to generate."""
    max_length: int = 1024
    """Tokenizer maximum length."""
    truncation: bool = False
    """Tokenizer truncation"""
    padding: bool = False 
    """Padding truncation"""
    gpu_layers: int = 0
    """GGML """
    
    _model: Optional[AutoModelForCausalLM] = Field(default=None)
    """The LlamaForCausalLM instance for this model."""

    def __init__(
        self,
        model_name: str = None,
        model_file: str = None,
        gpu_layers: int = None,
        temperature: Optional[float] = None,
        model_kwargs: Optional[Dict[str, Any]] = None,
        max_length: Optional[int] = None,
        truncation: Optional[bool] = None,
        padding: Optional[bool] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)

        if model_name is not None:
            self.model_name = model_name

        if temperature is not None:
            self.temperature = temperature

        if model_kwargs is not None:
            self.model_kwargs = model_kwargs

        if max_length is not None:
            self.max_length = max_length

        if truncation is not None:
            self.truncation = truncation

        if padding is not None:
            self.padding = padding

        if model_file is not None:
            self.model_file = model_file
        
        if gpu_layers is not None:
            self.gpu_layers = gpu_layers

        print("Loading in GGML Model with GPU layers")

        self.model = AutoModelForCausalLM.from_pretrained(
                self.model_name,
                model_file=self.model_file,
                gpu_layers=self.gpu_layers
        )

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value):
        # You can add custom logic here if needed
        self._model = value
        
    def generate(
            self,
            messages: List[BaseMessage],
            functions: Optional[List[Tool]] = None,
            stop: Optional[List[str]] = None,
            planning: Optional[bool] = True
        ) -> LLMResult:
            prompt = self._construct_prompt_from_message(messages)
            print("Prompt:", prompt)

            # inputs = [msg.content for msg in messages]
            # Concatenate the inputs into a single string
            # concatenated_string = "".join(prompt)
            # Remove leading and trailing whitespaces
            # result = concatenated_string.strip()
            
            text = self.model(prompt)
        
            generation = [Generation(message=BaseMessage(content=text))]
            # Return the result
            return self._create_llm_result(generation=generation, prompt=prompt, stop=stop, planning=planning)

    class Config:
        """Configuration for this Pydantic object."""
        arbitrary_types_allowed = True
        extra = Extra.ignore
        validate_assignment = False  # Turn off Pydantic validation errors

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling the language model API."""
        return {
            "model": self.model_name,
            "request_timeout": self.request_timeout,
            "max_tokens": self.max_tokens,
            "n": self.n,
            "temperature": self.temperature,
            **self.model_kwargs,
        }

    def _create_retry_decorator(self) -> Callable[[Any], Any]:
        min_seconds = 1
        max_seconds = 60
        # Wait 2^x * 1 second between each retry starting with
        # 4 seconds, then up to 10 seconds, then 10 seconds afterward
        return retry(
            reraise=True,
            stop=stop_after_attempt(self.max_retries),
            wait=wait_exponential(multiplier=1, min=min_seconds, max=max_seconds),
            retry=(
                retry_if_exception_type(
                    Exception
                )  # Replace this with specific exception types from the provider.
            ),
            before_sleep=before_sleep_log(logger, logging.WARNING),
        )

    def generate_with_retry(self, **kwargs: Any) -> Any:
        """Use tenacity to retry the completion call."""
        retry_decorator = self._create_retry_decorator()

        @retry_decorator
        def _generate_with_retry(**kwargs: Any) -> Any:
            # Replace this with the actual generate function from the provider.
            return self.model.generate(**kwargs)

        return _generate_with_retry(**kwargs)
    
    @staticmethod
    def _construct_prompt_from_message(messages: List[BaseMessage]):
        prompt = ""
        for msg in messages:
            prompt += msg.content
        return prompt

    @staticmethod
    def _enforce_stop_tokens(text: str, stop: List[str]) -> str:
        """Cut off the text as soon as any stop words occur."""
        first_index = len(text)
        for s in stop:
            if s in text:
                first_index = min(text.index(s), first_index)

        return text[:first_index].strip()

    def _create_llm_result(
        self, generation: Any, prompt: str, stop: List[str], planning=False
    ) -> LLMResult:
        print("Generation: ", str(generation))
        print("Planning: ", planning)
         
        ai_response = generation[0].message.content
                  
        print("AI Response: ", ai_response)

        return LLMResult(
                generations=[Generation(message=AIMessage(content=ai_response))],
                llm_output={
                    "token_usage": len(str(generation).split()),
                    "model_name": self.model_name,
                },
        )


from termcolor import colored
from transformers import PreTrainedModel, PreTrainedTokenizer
import json

from typing import List, Union, Dict, Any

from termcolor import cprint
from transformers import PreTrainedModel, PreTrainedTokenizer
import json

GENERATION_MARKER = "|GENERATION|"

class Jsonformer:
    value: Dict[str, Any] = {}

    def __init__(
        self,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizer,
        json_schema: Dict[str, Any],
        prompt: str,
        *,
        debug: bool = False,
        max_array_length: int = 10,
        max_number_tokens: int = 1024,
        temperature: float = 0.7,
        max_string_token_length: int = 1024,
        load_in_4bit = False,
    ):
        self.model = model
        self.tokenizer = tokenizer
        self.json_schema = json_schema
        self.prompt = prompt

        self.number_logit_processor = OutputNumbersTokens(self.tokenizer, self.prompt)

        self.generation_marker = "|GENERATION|"
        self.debug_on = debug
        self.max_array_length = max_array_length

        self.max_number_tokens = max_number_tokens
        self.temperature = temperature
        self.max_string_token_length = max_string_token_length
        self.load_in_4bit = load_in_4bit

    def debug(self, caller: str, value: str, is_prompt: bool = False):
        if self.debug_on:
            if is_prompt:
                cprint(caller, "green", end=" ")
                cprint(value, "yellow")
            else:
                cprint(caller, "green", end=" ")
                cprint(value, "blue")

    def generate_number(self, temperature: Union[float, None] = None, iterations=0):
        prompt = self.get_prompt()
        self.debug("[generate_number]", prompt, is_prompt=True)
        input_tokens = self.tokenizer.encode(prompt, return_tensors="pt").to(
            self.model.device
        )
        response = self.model.generate(
            input_tokens,
            max_new_tokens=self.max_number_tokens,
            num_return_sequences=1,
            logits_processor=[self.number_logit_processor],
            stopping_criteria=[
                NumberStoppingCriteria(self.tokenizer, len(input_tokens[0]))
            ],
            temperature=temperature or self.temperature,
            pad_token_id=self.tokenizer.eos_token_id,
        )
        response = self.tokenizer.decode(response[0], skip_special_tokens=True)

        response = response[len(prompt) :]
        response = response.strip().rstrip(".")
        self.debug("[generate_number]", response)
        try:
            return float(response)
        except ValueError:
            if iterations > 3:
                raise ValueError("Failed to generate a valid number")

            return self.generate_number(temperature=self.temperature * 1.3, iterations=iterations+1)

    def generate_boolean(self) -> bool:
        prompt = self.get_prompt()
        self.debug("[generate_boolean]", prompt, is_prompt=True)

        input_tensor = self.tokenizer.encode(prompt, return_tensors="pt")
        output = self.model.forward(input_tensor.to(self.model.device))
        logits = output.logits[0, -1]

        # todo: this assumes that "true" and "false" are both tokenized to a single token
        # this is probably not true for all tokenizers
        # this can be fixed by looking at only the first token of both "true" and "false"
        true_token_id = self.tokenizer.convert_tokens_to_ids("true")
        false_token_id = self.tokenizer.convert_tokens_to_ids("false")

        result = logits[true_token_id] > logits[false_token_id]

        self.debug("[generate_boolean]", result)

        return result.item()

    def generate_string(self) -> str:
        prompt = self.get_prompt() + '"'
        self.debug("[generate_string]", prompt, is_prompt=True)
        input_tokens = self.tokenizer.encode(prompt, return_tensors="pt").to(
            self.model.device
        )

        response = self.model.generate(
            input_tokens,
            max_new_tokens=self.max_string_token_length,
            num_return_sequences=1,
            temperature=self.temperature,
            stopping_criteria=[
                StringStoppingCriteria(self.tokenizer, len(input_tokens[0]))
            ],
            pad_token_id=self.tokenizer.eos_token_id,
        )

        # Some models output the prompt as part of the response
        # This removes the prompt from the response if it is present
        if (
            len(response[0]) >= len(input_tokens[0])
            and (response[0][: len(input_tokens[0])] == input_tokens).all()
        ):
            response = response[0][len(input_tokens[0]) :]
        if response.shape[0] == 1:
            response = response[0]

        response = self.tokenizer.decode(response, skip_special_tokens=True)

        self.debug("[generate_string]", "|" + response + "|")

        if response.count('"') < 1:
            return response

        return response.split('"')[0].strip()

    def generate_object(
        self, properties: Dict[str, Any], obj: Dict[str, Any]
    ) -> Dict[str, Any]:
        for key, schema in properties.items():
            self.debug("[generate_object] generating value for", key)
            obj[key] = self.generate_value(schema, obj, key)
        return obj

    def generate_value(
        self,
        schema: Dict[str, Any],
        obj: Union[Dict[str, Any], List[Any]],
        key: Union[str, None] = None,
    ) -> Any:
        schema_type = schema["type"]
        if schema_type == "number":
            if key:
                obj[key] = self.generation_marker
            else:
                obj.append(self.generation_marker)
            return self.generate_number()
        elif schema_type == "boolean":
            if key:
                obj[key] = self.generation_marker
            else:
                obj.append(self.generation_marker)
            return self.generate_boolean()
        elif schema_type == "string":
            if key:
                obj[key] = self.generation_marker
            else:
                obj.append(self.generation_marker)
            return self.generate_string()
        elif schema_type == "array":
            new_array = []
            obj[key] = new_array
            return self.generate_array(schema["items"], new_array)
        elif schema_type == "object":
            new_obj = {}
            if key:
                obj[key] = new_obj
            else:
                obj.append(new_obj)
            return self.generate_object(schema["properties"], new_obj)
        else:
            raise ValueError(f"Unsupported schema type: {schema_type}")

    def generate_array(self, item_schema: Dict[str, Any], obj: Dict[str, Any]) -> list:
        for _ in range(self.max_array_length):
            # forces array to have at least one element
            element = self.generate_value(item_schema, obj)
            obj[-1] = element

            obj.append(self.generation_marker)
            input_prompt = self.get_prompt()
            obj.pop()
            input_tensor = self.tokenizer.encode(input_prompt, return_tensors="pt")
            output = self.model.forward(input_tensor.to(self.model.device))
            logits = output.logits[0, -1]


            top_indices = logits.topk(30).indices
            sorted_token_ids = top_indices[logits[top_indices].argsort(descending=True)]

            found_comma = False
            found_close_bracket = False

            for token_id in sorted_token_ids:
                decoded_token = self.tokenizer.decode(token_id)
                if ',' in decoded_token:
                    found_comma = True
                    break
                if ']' in decoded_token:
                    found_close_bracket = True
                    break

            if found_close_bracket or not found_comma:
                break

        return obj

    def get_prompt(self):
        template = """{prompt}\nOutput result in the following JSON schema format:\n{schema}\nResult: {progress}"""
        progress = json.dumps(self.value)
        gen_marker_index = progress.find(f'"{self.generation_marker}"')
        if gen_marker_index != -1:
            progress = progress[:gen_marker_index]
        else:
            raise ValueError("Failed to find generation marker")

        prompt = template.format(
            prompt=self.prompt,
            schema=json.dumps(self.json_schema),
            progress=progress,
        )

        return prompt

    def __call__(self) -> Dict[str, Any]:
        self.value = {}
        generated_data = self.generate_object(
            self.json_schema["properties"], self.value
        )
        return generated_data

def highlight_values(value):
    def recursive_print(obj, indent=0, is_last_element=True):
        if isinstance(obj, dict):
            print("{")
            last_key = list(obj.keys())[-1]
            for key, value in obj.items():
                print(f"{' ' * (indent + 2)}{key}: ", end="")
                recursive_print(value, indent + 2, key == last_key)
            print(f"{' ' * indent}}}", end=",\n" if not is_last_element else "\n")
        elif isinstance(obj, list):
            print("[")
            for index, value in enumerate(obj):
                print(f"{' ' * (indent + 2)}", end="")
                recursive_print(value, indent + 2, index == len(obj) - 1)
            print(f"{' ' * indent}]", end=",\n" if not is_last_element else "\n")
        else:
            if isinstance(obj, str):
                obj = f'"{obj}"'
            print(colored(obj, "green"), end=",\n" if not is_last_element else "\n")

    recursive_print(value)

from typing import List
from transformers import PreTrainedTokenizer, LogitsWarper, StoppingCriteria
import torch

class StringStoppingCriteria(StoppingCriteria):
    def __init__(self, tokenizer: PreTrainedTokenizer, prompt_length: int):
        self.tokenizer = tokenizer
        self.prompt_length = prompt_length

    def __call__(
        self,
        input_ids: torch.LongTensor,
        _,
    ) -> bool:
        if len(input_ids[0]) <= self.prompt_length:
            return False

        last_token_id = input_ids[0][-1]
        last_token = self.tokenizer.decode(last_token_id, skip_special_tokens=True)

        result = '"' in last_token

        return result


class NumberStoppingCriteria(StoppingCriteria):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        prompt_length: int,
        precision: int = 3,
    ):
        self.tokenizer = tokenizer
        self.precision = precision
        self.prompt_length = prompt_length

    def __call__(
        self,
        input_ids: torch.LongTensor,
        scores: torch.FloatTensor,
    ) -> bool:
        decoded = self.tokenizer.decode(
            input_ids[0][self.prompt_length :], skip_special_tokens=True
        )

        if decoded.count(".") > 1:
            return True

        if (
            decoded.count(".") == 1
            and len(decoded.strip().split(".")[1]) > self.precision
        ):
            return True

        if (
            len(decoded) > 1
            and any(c.isdigit() for c in decoded)
            and decoded[-1] in [" ", "\n"]
        ):
            return True

        return False

class OutputNumbersTokens(LogitsWarper):
    def __init__(self, tokenizer: PreTrainedTokenizer, prompt: str):
        self.tokenizer = tokenizer
        self.tokenized_prompt = tokenizer(prompt, return_tensors="pt")
        vocab_size = len(tokenizer)
        self.allowed_mask = torch.zeros(vocab_size, dtype=torch.bool)

        for _, token_id in tokenizer.get_vocab().items():
            token_str = tokenizer.decode(token_id).strip()

            if token_str == "" or (
                all(c.isdigit() or c == "." for c in token_str)
                and token_str.count(".") <= 1
            ):
                self.allowed_mask[token_id] = True

    def __call__(self, _, scores):
        mask = self.allowed_mask.expand_as(scores)
        scores[~mask] = -float("inf")

        return scores


from .deanonymize import Deanonymize
from .sensitive_pii import SensitivePII
from .factuality_detector.factuality_detector import FactualityDetector
from .stop_output_substrings import StopOutputSubstrings
from .contradiction import Contradictions
from .malware_url import MalwareOutputURL
from .toxicity import ToxicityOutput
from .harmful import HarmfulOutput
from .text_quality import TextQualityOutput
from .coding_language import CodingLanguageOutput
from .relevance import Relevance
from .regex import RegexOutput
from .bias import Bias
from .factuality_detector.hallucination_kb_detector import HallucinationKBDetector

__all__ = [
    "Deanonymize",
    "SensitivePII",
    "FactualityDetector",
    "StopOutputSubstrings",
    "Contradictions",
    "MalwareOutputURL",
    "ToxicityOutput",
    "HarmfulOutput", 
    "TextQualityOutput",
    "CodingLanguageOutput"
    "Relevance",
    "RegexOuput",
    "Bias",
    "HallucinationKBDetector"
]


from st_aggrid import AgGrid
from st_aggrid.grid_options_builder import GridOptionsBuilder
from st_aggrid.shared import GridUpdateMode, JsCode

MAX_TABLE_HEIGHT = 500


def get_numeric_style_with_precision(precision: int) -> dict:
    return {"type": ["numericColumn", "customNumericFormat"], "precision": precision}


PRECISION_ZERO = get_numeric_style_with_precision(0)
PRECISION_ONE = get_numeric_style_with_precision(1)
PRECISION_TWO = get_numeric_style_with_precision(2)
PINLEFT = {"pinned": "left"}


def draw_grid(
    df,
    formatter: dict = None,
    selection="multiple",
    use_checkbox=False,
    fit_columns=False,
    theme="streamlit",
    max_height: int = MAX_TABLE_HEIGHT,
    wrap_text: bool = False,
    auto_height: bool = False,
    grid_options: dict = None,
    key=None,
    css: dict = None,
):
    gb = GridOptionsBuilder()
    gb.configure_default_column(
        filterable=True,
        groupable=False,
        editable=False,
        wrapText=wrap_text,
        autoHeight=auto_height,
    )

    if grid_options is not None:
        gb.configure_grid_options(**grid_options)

    for latin_name, (cyr_name, style_dict) in formatter.items():
        gb.configure_column(latin_name, header_name=cyr_name, **style_dict)

    gb.configure_selection(selection_mode=selection, use_checkbox=use_checkbox)

    return AgGrid(
        df,
        gridOptions=gb.build(),
        update_mode=GridUpdateMode.SELECTION_CHANGED | GridUpdateMode.VALUE_CHANGED,
        allow_unsafe_jscode=True,
        fit_columns_on_grid_load=fit_columns,
        height=max(max_height, (1 + len(df.index)) * 29),
        theme=theme,
        key=key,
        custom_css=css,
    )


def highlight(color, condition):
    code = f"""
        function(params) {{
            color = "{color}";
            if ({condition}) {{
                return {{
                    'backgroundColor': color
                }}
            }}
        }};
    """
    return JsCode(code)



import time
import json
import requests
import openai
from threading import Thread

from guardrail.metrics.utils.keys import init_keys, load_api_key
from guardrail.env_config import guardrail_env_config
from guardrail.metrics.utils.flatten_metrics import flatten_json
from guardrail.metrics.offline_evals.offline_metrics import OfflineMetricsEvaluator

class OpenAI:
    def __init__(self, firewall, llm_model="gpt-3.5-turbo", api_keys={}):
        init_keys()
        self.llm_model = llm_model
        self.offline_metrics = OfflineMetricsEvaluator()
        self.firewall = firewall

    def create_completion(self, engine, prompt, gr_tags=None, **kwargs):
        kwargs['engine'] = engine
        kwargs['prompt'] = prompt
        if gr_tags:
            kwargs['gr_tags'] = gr_tags
        result = openai.Completion.create(**kwargs)        

    def create_chat_completion(self, model, messages, **kwargs):
        kwargs['model'] = model
        kwargs['messages'] = messages

        start_time = time.time()
        openai_response = self._chat_completion_request(**kwargs)
        
        prompt = self.get_latest_user_prompt(messages)
        chatbot_response = openai_response["choices"][0]["message"]["content"]

        end_time = time.time()
        total_time = end_time - start_time
        print(f"ChatCompletion executed in {total_time:.4f} seconds")

        return openai_response, chatbot_response
     
    def get_latest_user_prompt(self, messages):
        for message in reversed(messages):
            if message["role"] == "user":
                return message["content"]
        return None

    def run_chat_completion(self, 
                            messages, 
                            model="gpt-3.5-turbo", 
                            temperature=0, 
                            max_tokens=256, 
                            **kwargs):
        if isinstance(messages, str):
            messages = [{'role': 'user', 'content': messages}]
        
        start_time = time.time()

        prompt = messages[-1]['content']

        if self.firewall == None:
            openai_response, chatbot_response = self.create_chat_completion(
                model=model,
                messages=messages,
                temperature=temperature,
                max_tokens=max_tokens,
                **kwargs
            )
        else:
            sanitized_prompt, input_valid_results, input_risk_score = self.firewall.scan_input(prompt)
            print("Prompt Valid Results: ", input_valid_results)
            openai_response, chatbot_response = self.create_chat_completion(
                model=model,
                messages=messages,
                temperature=temperature,
                max_tokens=max_tokens,
                **kwargs
            )

            sanitized_response, output_valid_results, output_risk_score = self.firewall.scan_output(
                sanitized_prompt, chatbot_response
            )
            print("Output Valid Results: ", output_valid_results)

        end_time = time.time()
        total_time = end_time - start_time

        # Start storing logs in the background
        # eval_log_thread = Thread(target=self.run_eval_store_logs, args=(prompt, chatbot_response, openai_response["usage"], total_time, input_risk_score, output_risk_score))
        # eval_log_thread.start()

        # eval_log_thread.join()
        print(f"Total time executed in {total_time:.4f} seconds")

        if self.firewall == None:
            self.run_eval_store_logs(prompt, chatbot_response, openai_response["usage"], total_time, {}, {})
            return openai_response
        else:
            self.run_eval_store_logs(prompt, chatbot_response, openai_response["usage"], total_time, input_risk_score, output_risk_score)
            return sanitized_response, input_risk_score, output_risk_score



    def run_chat_completion_entry(self, user_message_content, model="gpt-3.5-turbo", temperature=0, max_tokens=256, **kwargs):
        self.run_chat_completion(user_message_content, model, temperature, max_tokens, **kwargs)

    def _chat_completion_request(self, **kwargs):
        return openai.ChatCompletion.create(**kwargs)

    def convert_numbers_to_strings(self, d):
        new_dict = {}
        for key, value in d.items():
            if isinstance(value, dict):
                # If the value is another dictionary, recursively call the function
                new_dict[key] = self.convert_numbers_to_strings(value)
            elif isinstance(value, (int, float)):
                # If the value is an int or float, convert it to a string
                new_dict[key] = str(value)
            else:
                new_dict[key] = value  # Keep other types unchanged
        return new_dict

    def flatten_evaluations(self, evaluations, parent_key='', separator='_'):
        """
        Flatten the "evaluations" dictionary.

        Parameters:
        - evaluations (dict): The "evaluations" dictionary to flatten.
        - parent_key (str): The parent key used for recursion. Leave empty for the initial call.
        - separator (str): The separator to use between flattened keys.

        Returns:
        - dict: The flattened "evaluations" dictionary.
        """
        flattened = {}
        risk_flag = "pass"
        failure_reasons = []  # Initialize an empty list for failure reasons
        for key, value in evaluations.items():
            new_key = f"{parent_key}{separator}{key}" if parent_key else key
            if isinstance(value, dict):
                result = self.flatten_evaluations(value, new_key, separator)
                flattened.update(result)
                if result.get("risk_status") == "fail":
                    failure_reasons.append(result.get("failure_reason"))
            else:
                flattened[new_key] = value
                try:
                    float_value = float(value)
                    if key == "Anonymize" and float_value >= 0.75:
                        risk_flag = "mitigated"
                    elif 0.75 <= float_value <= 1.0:
                        risk_flag = "fail"
                        # Add the triggering key to the failure reasons list
                        failure_reasons.append(new_key)
                except ValueError:
                    # Handle the case where the string is not a valid float
                    pass
        flattened["risk_status"] = risk_flag
        if risk_flag == "fail":
            flattened["failure_reason"] = " ".join(failure_reasons)
        return flattened

    def run_eval_store_logs(self, prompt, output, openai_usage, total_time, input_risk_score, output_risk_score, model_name="gpt-3.5-turbo"):
        # Simulate storing logs
        print("prompt: ", prompt)
        print("output: ", output)
        cost = self.calculate_openai_cost(openai_usage, model_name)

        promptEvaluations = self.flatten_evaluations(self.convert_numbers_to_strings(input_risk_score))
        outputEvaluations = self.flatten_evaluations(self.convert_numbers_to_strings(output_risk_score))

        total_tokens = openai_usage["total_tokens"]

        api_key = guardrail_env_config.guardrail_api_key
        if api_key == "" or api_key == None:
            api_key = load_api_key("GUARDRAIL_API_KEY")
    
        api_url = "https://guardrail-api-sknmgpkina-uc.a.run.app/v1/create_evaluation_log"
        headers = {
            'Content-Type': 'application/json',  # Set Content-Type to indicate JSON data
            'Authorization': f'Bearer {api_key}'  # Replace api_key with your actual token
        }

        data = {
            "prompt": prompt,
            "output": output,
            "cost": str(cost),
            "llmToken": str(total_tokens),
            "latency": str(total_time),
            "promptEvaluations": promptEvaluations,
            "outputEvaluations": outputEvaluations,
        }

        response = requests.post(api_url, headers=headers, json=data)
        print(response.status_code)

        if response.status_code == 200:
            print("Stored logs successfully: ", response)
            print(response.json())
            return response.json()
        elif response.status_code == 403 or response.status_code >= 400:
            # Capture 403 errors and any 400+ errors
            print("Error Status Code:", response.status_code)
            print("Response Content:", response.content)
            print("Response Headers:", response.headers)
            return None
        else:
            # Handle other status codes as needed
            print("Unhandled Status Code:", response.status_code)
            return None

    def _run_offline_metrics(self, prompt, response):
        results = self.offline_metrics.evaluate_metrics(response, prompt)
        results = flatten_json(results)
        print(results)
        return results

    def calculate_openai_cost(self, usage, model="gpt-3.5-turbo"):
        pricing = {
            'gpt-3.5-turbo': {
                'prompt': 0.0015,
                'completion': 0.002,
            },
            'gpt-3.5-turbo-16k': {
                'prompt': 0.003,
                'completion': 0.004,
            },
            'gpt-4-8k': {
                'prompt': 0.03,
                'completion': 0.06,
            },
            'gpt-4-32k': {
                'prompt': 0.06,
                'completion': 0.12,
            },
            'text-embedding-ada-002-v2': {
                'prompt': 0.0001,
                'completion': 0.0001,
            }
        }

        try:
            model_pricing = pricing[model]
        except KeyError:
            raise ValueError("Invalid model specified")

        prompt_cost = usage['prompt_tokens'] * model_pricing['prompt'] / 1000
        completion_cost = usage['completion_tokens'] * model_pricing['completion'] / 1000

        total_cost = prompt_cost + completion_cost
        print(f"\nTokens used:  {usage['prompt_tokens']:,} prompt + {usage['completion_tokens']:,} completion = {usage['total_tokens']:,} tokens")
        print(f"Total cost for {model}: ${total_cost:.6f}\n")

        return total_cost


class CustomMetricsWrapper:
    def __init__(self, custom_metric_function):
        self.custom_metric_function = custom_metric_function

    def evaluate(self, *args, **kwargs):
        return self.custom_metric_function(*args, **kwargs)



class OutputParserException(Exception):
    """Exception raised by output parsers to signify a parsing error.

    This exception is specifically used for handling parsing errors encountered while processing
    the output from a tool. It allows differentiation between parsing errors and other code or
    execution errors that may occur within the output parser. When an OutputParserException is
    raised, it can be caught and handled in a way to address the parsing error, while other errors
    will be raised as usual.
    """

    pass


class ToolRunningError(Exception):
    """Exception raised when a tool fails to run.

    This exception is raised to indicate that an error occurred while attempting to run a tool. It
    can be caught and handled by the application to address the specific failure in running the tool.
    The `message` attribute of this exception contains information about the reason for the failure.
    """

    def __init__(self, message):
        self.message = message



import json
import csv

instruction = "You will answer questions about the US constitution "

with open("./responses.json", "r") as f:
    responses = json.load(f)

# Open the CSV file and write the data to it
with open("responses.csv", "w", newline="") as csvfile:
    fieldnames = ["question", "answer"]
    writer = csv.DictWriter(csvfile, fieldnames=fieldnames)

    writer.writeheader()
    for response in responses["responses"]:
        if "question" in response and "answer" in response:  # To make sure both keys exist
            writer.writerow(
                {
                    "question": instruction + response["question"],
                    "answer": response["answer"],
                }
            )



import PyPDF2
from transformers import AutoTokenizer, LlamaTokenizer, pipeline
import json
import requests
import re
import torch
import csv
from cleantext import clean
from unstructured.partition.auto import partition

# 1. Change Tokenizer
# 2. Change system_template
# 3. Change PDF file


class DatasetCreator:
    def __init__(
        self,
        file_path=None,
        model=None,
        model_endpoint=None,
        model_pipeline=None,
        tokenizer=None,
        prompt=None,
        output_filetype=None,
        system=None,
        instruction=None,
    ):
        self.file_path = file_path
        self.model = model
        self.model_endpoint = (model_endpoint,)
        self.model_pipeline = (model_pipeline,)
        self.tokenizer = tokenizer
        self.prompt = prompt
        self.output_filetype = output_filetype
        self.system = system
        self.instruction = instruction
        self.text_gen = pipeline(
            model=self.model,
            tokenizer=self.tokenizer,
            torch_dtype=torch.bfloat16,
            trust_remote_code=True,
            load_in_8bit=True,
            device_map="auto",
        )

    def create_dataset(self):
        self.initialize_vars()
        token = False
        if self.model_pipeline:
            chunks = self.split_text(self.file_path)
        else:
            chunks = self.tokenize_file(self.file_path)
            token = True
        responses = self.get_responses(chunks, token, [])
        self.convert_json_output(responses)
        print("Completed creating output file")

    def initialize_vars(self):
        if (
            self.file_path is None
            or self.model is None
            or self.tokenizer is None
            or self.output_filetype is None
        ):
            print("Missing required arguments.")
            return

        if self.model is None and self.model_endpoint is None:
            print("Missing either model or endpoint to run inference")
            return

        if self.prompt is None:
            self.prompt = (
                "You are an API that converts bodies of text into a single question and answer into a JSON format. Each JSON "
                "contains a single question with a single answer. Only respond with the JSON and no additional text. \n"
            )

        if self.system is None:
            self.system = "You are an AI assistant that follows system extremely well. Help as much as you can."

        if self.output_filetype is None:
            self.output_filetype = "csv"

        if self.instruction is None:
            self.instruction = "You will answer questions about the files"

    # TODO support local oobabooga inference API
    def generate_qa_pair(self, system, prompt, input):
        if input:
            # prompt = f"### System:\n{system}\n\n### User:\n{prompt}\n\n### Input:\n{input}\n\n### Response:\n"
            prompt = self.prompt + input
        else:
            prompt = f"### System:\n{system}\n\n### User:\n{prompt}\n\n### Response:\n"

        if self.text_gen:
            instance = {
                "top_p": 0.3,
                "temperature": 0.4,
                "generate_len": 256,
                "top_k": 10,
            }

            output = self.text_gen(
                prompt,
                top_k=instance["top_k"],
                use_cache=True,
                do_sample=True,
                top_p=instance["top_p"],
                temperature=instance["temperature"],
            )

            return output[0]["generated_text"].replace("\n", "")

        else:
            instance = {
                "input_ids": tokens,
                "top_p": 0.2,
                "temperature": 0.3,
                "generate_len": 256,
                "top_k": 3,
            }

            tokens = self.tokenizer.encode(prompt)
            tokens = torch.LongTensor(tokens).unsqueeze(0)
            tokens = tokens.to("cuda")

            length = len(tokens[0])

            with torch.no_grad():
                rest = self.model.generate(
                    input_ids=tokens,
                    max_length=length + instance["generate_len"],
                    use_cache=True,
                    do_sample=True,
                    top_p=instance["top_p"],
                    temperature=instance["temperature"],
                    top_k=instance["top_k"],
                )
            output = rest[0][length:]
            string = self.tokenizer.decode(output, skip_special_tokens=True)
            return f"[!] Response: {string}"

    """
    def extract_text_from_pdf(self, file_path):
        pdf_file_obj = open(file_path, 'rb')
        pdf_reader = PyPDF2.PdfReader(pdf_file_obj)
        text = ''
        for page_num in range(len(pdf_reader.pages)):
            page_obj = pdf_reader.pages[page_num]
            text += page_obj.extract_text()
        pdf_file_obj.close()
        cleaned_text = clean(text, lower=False)
        return cleaned_text
    """

    # TODO: PDF, HTML, DOCX
    def extract_text_from_unstructured(self, file_path):
        elements = partition(filename=file_path)
        text = "\n\n".join([str(el) for el in elements])
        cleaned_text = clean(text, lower=False)
        return cleaned_text

    def tokenize_pdf(self, path):
        text = self.extract_text_from_unstructured(path)
        tokens = self.tokenize(text)

        token_chunks = list(self.chunks(tokens, 256))
        print("Number of token_chunks", len(token_chunks))
        return token_chunks

    def tokenize(self, text):
        enc = self.tokenizer.encode(text)
        return enc

    def chunks(self, lst, n):
        for i in range(0, len(lst), n):
            yield lst[i : i + n]

    def is_json(self, data):
        try:
            json_data = json.loads(data)
            if "question" in json_data and "answer" in json_data:
                return True
        except ValueError:
            return False

    def convert_to_json(self, output):
        # Remove leading/trailing whitespaces and newlines
        output = output.strip()

        # Initialize an empty JSON object
        json_object = {}

        # Extract the JSON string from the output
        json_start = output.find("{")
        json_end = output.rfind("}")
        json_string = output[json_start : json_end + 1]

        # Try parsing the JSON string
        try:
            json_data = json.loads(json_string)
            if "question" in json_data and "answer" in json_data:
                json_object["question"] = json_data["question"]
                json_object["answer"] = json_data["answer"]
        except ValueError:
            # Extract question and answer from non-JSON format
            lines = output.splitlines()
            for line in lines:
                line = line.strip()
                if (
                    line.startswith("Response:")
                    or line.startswith('"response"')
                    or line.startswith("[!] Response:")
                ):
                    line = line.split(":", 1)[1].strip().strip('"').strip()
                if (
                    line.startswith("Question:")
                    or line.startswith('"question"')
                    or line.startswith("question:")
                ):
                    question = line.split(":", 1)[1].strip().strip('"').strip()
                    json_object["question"] = question
                elif line.startswith("Answer:") or line.startswith('"answer"'):
                    answer = line.split(":", 1)[1].strip().strip('"').strip()
                    json_object["answer"] = answer

        return json.dumps(json_object)

    def submit_to_api(self, chunk, retries=5):
        for i in range(retries):
            try:
                if i > 1:
                    initial_response = self.generate_qa_pair(
                        self.system, self.prompt, chunk.strip()
                    )
                initial_response = self.generate_qa_pair(self.system, self.prompt, chunk.strip())
                print("Initial response", initial_response)
                # Extract JSON string from between back-ticks
                response = self.convert_to_json(initial_response)
                print("cleaned up response", response)
                if self.is_json(response):
                    print(response)
                    return json.loads(response)
                else:
                    match = re.search(r"`(.*?)`", response, re.S)
                    if match and self.is_json(match.group(1)):
                        print(f"Attempt {i + 1} failed. Retrying...")
                        return json.loads(
                            match.group(1)
                        )  # assuming you want to return the JSON data
            except requests.exceptions.RequestException as e:
                print(f"Request failed: {e}")
                continue
        print("Max retries exceeded. Skipping this chunk.")
        return None

    def split_text(self, path):
        text = self.extract_text_from_unstructured(path)
        token_chunks = list(self.chunks(text, 256))
        print("Number of text chunks", len(token_chunks))
        return token_chunks

    def tokenize_file(self, path):
        text = self.extract_text_from_unstructured(path)
        tokens = self.tokenize(text)

        token_chunks = list(self.chunks(tokens, 256))
        print("Number of token_chunks", len(token_chunks))
        return token_chunks

    def get_responses(self, token_chunks, token, responses):
        for chunk in token_chunks:
            if token:
                response = self.submit_to_api(self.tokenizer.decode(chunk))
            else:
                response = self.submit_to_api(chunk)
            if response is not None:
                responses.append(response)
            print("Size of responses", len(responses))

        # Write responses to a JSON file
        with open("responses.json", "w") as f:
            json.dump(responses, f)

    def convert_json_output(self, responses):
        # Open the CSV file and write the data to it
        with open("responses.csv", "w", newline="") as csvfile:
            fieldnames = ["question", "answer"]
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)

            writer.writeheader()
            for response in responses["responses"]:
                if "question" in response and "answer" in response:  # To make sure both keys exist
                    writer.writerow(
                        {
                            "question": self.instruction + response["question"],
                            "answer": response["answer"],
                        }
                    )



import json
from abc import abstractmethod
from typing import Any, Dict, Union, List

from pydantic import BaseModel

from guardrail.guardchain.models.base_model import Generation, BaseLanguageModel
from guardrail.guardchain.agent.message import BaseMessage, UserMessage
from guardrail.guardchain.chain import constants
from guardrail.guardchain.errors import OutputParserException


class AgentAction(BaseModel):
    """Agent's action to take."""

    tool: str
    tool_input: Union[str, dict]
    tool_output: str = ""
    log: str = ""
    model_response: str = ""

    @property
    def response(self) -> str:
        """Get the message to be stored in memory and shared with the next prompt."""
        if self.model_response and not self.tool_output:
            return self.model_response
        return (
            f"Outputs from using tool '{self.tool}' for inputs {self.tool_input} "
            f"is '{self.tool_output}'\n"
        )


class AgentFinish(BaseModel):
    """Agent's return value."""

    message: str
    log: str
    intermediate_steps: List[AgentAction] = []

    def format_output(self) -> Dict[str, Any]:
        final_output = {
            "message": self.message,
            constants.INTERMEDIATE_STEPS: self.intermediate_steps,
        }
        return final_output


class AgentOutputParser(BaseModel):
    @staticmethod
    def load_json_output(message: BaseMessage, llm) -> Dict[str, Any]:
        """If the message contains a json response, try to parse it into dictionary"""
        text = message.content
        clean_text = ""
        
        try:
            clean_text = text[text.index("{") : text.rindex("}") + 1].strip()
            response = json.loads(clean_text)
        except Exception:
            message = [
                UserMessage(
                    content=f"""Fix the following json into correct format
                                ```json
                                {clean_text}
                                ```
                                """
                )
            ]
            full_output: Generation = llm.generate(message).generations[0]
            response = json.loads(full_output.message.content)

        return response

    @abstractmethod
    def parse(self, message: BaseMessage) -> Union[AgentAction, AgentFinish]:
        """Parse text into agent action/finish."""

    def parse_clarification(
        self, message: BaseMessage, agent_action: AgentAction, llm
    ) -> Union[AgentAction, AgentFinish]:
        """Parse clarification outputs"""
        return agent_action



from __future__ import annotations

SAFE_PROMPT_TEMPLATE = """
{{#system~}}
You are a helpful, respectful and honest assistant. Always answer as helpfully as possible, while being safe.  Your answers should not include any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content. 
Please ensure that your responses are socially unbiased and positive in nature.If a question does not make any sense, or is not factually coherent, explain why instead of answering something not correct. 

If you don't know the answer to a question, please don't share false information.
{{~/system}}
{{#user~}}
I want a response to the most recent query in the conversation history, while accounting for the history for context:

Conversation History:
{{history}}

Is the question free from any content that encourages violence, harm to others, explicit or inappropriate material, hate speech, discrimination, self-harm, illegal activities, harassment, bullying, malicious intent, explicit sexual content,
or attempts to exploit or manipulate the language model for inappropriate purposes?

Please don't answer the question or comment on it yet. Answer in JSON "safe_prompt" key from valid_options = ["yes", "no"], probability distribution of likelihood of each option, and a rationale key if "response" key is "no".

Example JSON format:
```json
{
  "safe_prompt": "gen 'safe_prompt'",
  "probability_distribution": {
    "yes": [0.00 - 1.00],
    "no": [0.00 - 1.00]
  }
}
```
{{~/user}}
{{#assistant~}}
{{gen 'experts' temperature=0 max_tokens=300}}
{{~/assistant}}
"""

PLANNING_PROMPT_TEMPLATE = """
{{#system~}}
A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions.

You are a helpful, respectful and honest assistant who tries to have helpful conversation with the 
user based on previous conversation and previous tools outputs from tools. Here's some tools and context:

Assistant has access to the following tools:
{{tools}}

Previous tools outputs:
{{agent_scratchpad}}

Conversation History:
{{history}}

Please respond to user prompt in JSON format as described below:
RESPONSE FORMAT (FOR CONTEXT ONLY, BUT FOLLOW JSON STRUCTURE):
{
  "thoughts": {
    "plan": "Given previous tools outputs, what is the next step after the previous conversation",
    "need_use_tool": "answer with 'Yes' if requires more information not in previous tools outputs else 'No'"
  },
  "tool": {
    "name": "tool name, should be one of [${tool_names}] or empty if tool is not needed",
    "args": {
      "arg_name": "arg value from conversation history or tools outputs to run tool"
    }
  },
  "response": "response to user given tools outputs and conversations",
}

Ensure the response can be parsed by Python json.loads
{{~/system}}

{{#user~}}
I want a response to the following question:

{{query}}}
{{~/user}}

{{#assistant~}}
{{gen 'planning_prompt' temperature=0 max_tokens=300}}
{{~/assistant}}

"""

SHOULD_ANSWER_PROMPT_TEMPLATE = """
A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions. 

Conversation History:
${history}

USER:
Given the following conversation so far, have you finished helping me (user) with all the 
questions and prompts? Answer with "yes" or "no" based on conversation.

ASSISTANT:
"""


PLANNING_PROMPT_TEMPLATE_ORIGINAL = """
### Instruction
You are a helpful, respectful and honest assistant who tries to have helpful conversation with the 
user based on previous conversation and previous tools outputs from tools. 

Always answer as helpfully as possible, while being safe.  
Your answers should not include any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content. 
Please ensure that your responses are socially unbiased and positive in nature.

${prompt}
Use tool when provided. If there is no tool available, respond with have a helpful and polite 
conversation. Find next step without using the same tool with same inputs.

Assistant has access to the following tools:
${tools}

### Context
Previous conversation so far:
${history}

Previous tools outputs:
${agent_scratchpad}

### Input:
${human_input}

Generate a new response to

### Assistant:
"""

SHOULD_ANSWER_PROMPT_TEMPLATE_ORIGINAL = """
### Instruction
A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions.

You are a helpful, respectful and honest assistant who tries to have helpful conversation with the 
user based on previous conversation and previous tools outputs from tools. 

Conversation History:
${history}

USER:
${human_input}

ASSISTANT:
"""

FIX_TOOL_INPUT_PROMPT_TEMPLATE = """Tool have the following spec and input provided
Spec: "{tool_description}"
Inputs: "{inputs}"
Running this tool failed with the following error: "{error}"
What is the correct input in JSON format for this tool?
"""


CLARIFYING_QUESTION_PROMPT_TEMPLATE = """You are a support agent who is going to use '${tool_name}' tool.
Check if you have enough information from the previous conversation and tools outputs to use tool based on the spec below.
"${tool_desp}"

Previous conversation so far:
${history}

Previous tools outputs:
${agent_scratchpad}

Please respond user question in JSON format as described below
RESPONSE FORMAT:
{
    "has_arg_value": "Do values for all input args for '${tool_name}' tool exist? answer with Yes or No",
    "clarifying_question": "clarifying question to user to ask for missing information"
}
Ensure the response can be parsed by Python json.loads"""


import logging
from textstat import textstat

class Textstat:
    def __init__(self):
        self.ts = textstat
        self.logger = logging.getLogger("Textstat")
        logging.basicConfig(level=logging.INFO)

    def evaluate(self, text):
        metrics = {
            "automated_readability_index": {
                "col_type": "String",
                "schema_name": None,
                "function": lambda text: str(self.ts.automated_readability_index(text)),
            },
            "dale_chall_readability_score": {
                "col_type": "String",
                "schema_name": "text_standard_component",
                "function": lambda text: str(self.ts.dale_chall_readability_score(text)),
            },
            "linsear_write_formula": {
                "col_type": "String",
                "schema_name": "text_standard_component",
                "function": lambda text: str(self.ts.linsear_write_formula(text)),
            },
            "gunning_fog": {
                "col_type": "String",
                "schema_name": "text_standard_component",
                "function": lambda text: str(self.ts.gunning_fog(text)),
            },
            "aggregate_reading_level": {
                "col_type": "String",
                "schema_name": None,
                "function": lambda text: str(self.ts.text_standard(text, float_output=True)),
            },
            "fernandez_huerta": {
                "col_type": "String",
                "schema_name": "es",
                "function": lambda text: str(self.ts.fernandez_huerta(text)),
            },
            "szigriszt_pazos": {
                "col_type": "String",
                "schema_name": "es",
                "function": lambda text: str(self.ts.szigriszt_pazos(text)),
            },
            "gutierrez_polini": {
                "col_type": "String",
                "schema_name": "es",
                "function": lambda text: str(self.ts.gutierrez_polini(text)),
            },
            "crawford": {
                "col_type": "String",
                "schema_name": "es",
                "function": lambda text: str(self.ts.crawford(text)),
            },
            "gulpease_index": {
                "col_type": "String",
                "schema_name": "it",
                "function": lambda text: str(self.ts.gulpease_index(text)),
            },
            "osman": {
                "col_type": "String",
                "schema_name": "ar",
                "function": lambda text: str(self.ts.osman(text)),
            },
            "flesch_kincaid_grade": {
                "col_type": "String",
                "schema_name": "text_standard_component",
                "function": lambda text: str(self.ts.flesch_kincaid_grade(text)),
            },
            "flesch_reading_ease": {
                "col_type": "String",
                "schema_name": None,
                "function": lambda text: str(self.ts.flesch_reading_ease(text)),
            },
            "smog_index": {
                "col_type": "String",
                "schema_name": "text_standard_component",
                "function": lambda text: str(self.ts.smog_index(text)),
            },
            "coleman_liau_index": {
                "col_type": "String",
                "schema_name": "text_standard_component",
                "function": lambda text: str(self.ts.coleman_liau_index(text)),
            },
            "sentence_count": {
                "col_type": "String",
                "schema_name": None,
                "function": lambda text: str(self.ts.sentence_count(text)),
            },
            "character_count": {
                "col_type": "String",
                "schema_name": None,
                "function": lambda text: str(self.ts.char_count(text)),
            },
            "letter_count": {
                "col_type": "String",
                "schema_name": None,
                "function": lambda text: str(self.ts.letter_count(text)),
            },
            "polysyllable_count": {
                "col_type": "String",
                "schema_name": None,
                "function": lambda text: str(self.ts.polysyllabcount(text)),
            },
            "monosyllable_count": {
                "col_type": "String",
                "schema_name": None,
                "function": lambda text: str(self.ts.monosyllabcount(text)),
            },
            "difficult_words": {
                "col_type": "String",
                "schema_name": None,
                "function": lambda text: str(self.ts.difficult_words(text)),
            },
            "syllable_count": {
                "col_type": "String",
                "schema_name": None,
                "function": lambda text: str(self.ts.syllable_count(str(text))),
            },
            "lexicon_count": {
                "col_type": "String",
                "schema_name": None,
                "function": lambda text: str(self.ts.lexicon_count(text)),
            },
        }

        results = {}
        for metric, config in metrics.items():
            try:
                result = config["function"](text)
                self.logger.info(f"{metric}: {result}")
                results[metric] = result
            except Exception as e:
                self.logger.error(f"Error while evaluating {metric}: {e}")
                results[metric] = None

        return results



import inspect
from abc import ABC
from typing import Any, Callable, Dict, Optional, Tuple, Type, Union

from pydantic import BaseModel, validator

from guardrail.guardchain.errors import ToolRunningError


class Tool(ABC, BaseModel):
    """Interface GuardChain tools must implement."""

    name: Optional[str]
    """The unique name of the tool that clearly communicates its purpose.
    If not provided, it will be named after the func name.
    The more descriptive it is, the easier it would be for the model to call the right tool.
    """

    description: str
    """Used to tell the model how/when/why to use the tool.
    You can provide few-shot examples as a part of the description.
    """

    arg_description: Optional[Dict[str, Any]] = None
    """Dictionary of arg name and description when using GuardChainAgent to provide 
    additional argument information."""

    args_schema: Optional[Type[BaseModel]] = None
    """Pydantic model class to validate and parse the tool's input arguments."""

    func: Union[Callable[..., str], None] = None

    @validator("name", pre=True)
    def validate_environment(cls, name, values):
        """Validate that api key and python package exist in the environment."""
        func = values.get("func")
        if func and not name:
            name = func.__name__

        # Check if all args from arg_description exist in func args
        if values.get("arg_description") and func:
            inspection = inspect.getfullargspec(func)
            override_args = set(values["arg_description"].keys())
            args = set(inspection.args)
            override_without_args = override_args - args
            if len(override_without_args) > 0:
                raise ValueError(
                    f"Provide arg description for not existed args: {override_without_args}"
                )

        return name

    def _parse_input(
        self,
        tool_input: Union[str, Dict],
    ) -> Union[str, Dict[str, Any]]:
        """Convert tool input to pydantic model."""
        input_args = self.args_schema
        if isinstance(tool_input, str):
            if input_args is not None:
                key_ = next(iter(input_args.__fields__.keys()))
                input_args.validate({key_: tool_input})
            return tool_input
        else:
            if input_args is not None:
                result = input_args.parse_obj(tool_input)
                return {k: v for k, v in result.dict().items() if k in tool_input}
        return tool_input

    def _to_args_and_kwargs(self, tool_input: Union[str, Dict]) -> Tuple[Tuple, Dict]:
        # For backwards compatibility, if run_input is a string,
        # pass it as a positional argument.
        if isinstance(tool_input, str):
            return (tool_input,), {}
        else:
            return (), tool_input

    def _run(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> str:
        return self.func(*args, **kwargs)

    def run(
        self,
        tool_input: Union[str, Dict] = "",
        **kwargs: Any,
    ) -> str:
        """Run the tool."""
        try:
            parsed_input = self._parse_input(tool_input)
        except ValueError as e:
            # Return the exception as tool output
            raise ToolRunningError(message=f"Tool input args value Error: {e}") from e

        try:
            tool_args, tool_kwargs = self._to_args_and_kwargs(parsed_input)
            tool_output = self._run(*tool_args, **tool_kwargs)
        except (Exception, KeyboardInterrupt) as e:
            raise ToolRunningError(message=f"Failed to run tool {self.name} due to {e}") from e

        return tool_output



import numpy as np
import umap
from sklearn.manifold import TSNE

class EmbeddingModel:
    def __init__(self, embedding_data, non_linear_function='sigmoid'):
        # 'embedding_data' is a dictionary mapping text data to its corresponding embeddings
        # For example: {'get python': np.array([0.8, 0.2, 0.5]), 'HTTP method': np.array([0.9, 0.3, 0.6]), ...}
        self.embedding_data = embedding_data

        # Define the non-linear scoring function based on the user's choice
        if non_linear_function == 'sigmoid':
            self.non_linear_function = self.sigmoid
        elif non_linear_function == 'relu':
            self.non_linear_function = self.relu
        elif non_linear_function == 'tanh':
            self.non_linear_function = self.tanh
        else:
            raise ValueError(f"Invalid non-linear function: {non_linear_function}. Supported functions are 'sigmoid', 'relu', and 'tanh'.")

    def sigmoid(self, x):
        return 1 / (1 + np.exp(-x))

    def relu(self, x):
        return np.maximum(0, x)

    def tanh(self, x):
        return np.tanh(x)

    def calculate_similarity(self, embedding1, embedding2):
        # Assuming 'embedding1' and 'embedding2' are NumPy arrays representing the embeddings of two results
        # Calculate the similarity between two embeddings (e.g., using cosine similarity)
        return np.dot(embedding1, embedding2) / (np.linalg.norm(embedding1) * np.linalg.norm(embedding2))

    def generate_freshness_score(self, embedding, weight=1.5, bias=-0.5):
        # Apply the selected non-linear transformation on the embedding

        transformed_score = weight * embedding + bias

        # Apply the selected non-linear function to map the transformed score
        freshness_score = self.non_linear_function(transformed_score)

        return freshness_score
    
    def calculate_similarity(self, embedding1, embedding2):
        # Assuming 'embedding1' and 'embedding2' are NumPy arrays representing the embeddings of two results
        # Calculate the similarity between two embeddings (e.g., using cosine similarity)
        return np.dot(embedding1, embedding2) / (np.linalg.norm(embedding1) * np.linalg.norm(embedding2))
    
    def improve_embeddings_with_diversity(self, query_embedding, diversity_weight=0.5, top_k=5):
        # Assuming 'query_embedding' is a NumPy array representing the embedding of the search query
        # 'diversity_weight' is a trade-off parameter between relevance and diversity (0.0 for pure relevance, 1.0 for pure diversity)
        # 'top_k' is the number of diverse results to select and return
        # Goal is to encourage the retrieval of not only the nearest neighbors but also diverse results that cover various aspects of the query's intent.
        # Maximal Marginal Relevance" (MMR), which balances relevance and diversity.
        # MMR selects results that are both relevant to the query and different from the ones already 
        # selected. MMR uses a trade-off parameter called the diversity weight to control the balance 
        # between relevance and diversity.

        # List to keep track of selected results
        selected_results = []

        # Select the most relevant result (nearest neighbor) as the first selected result
        result_texts = list(self.embedding_data.keys())
        similarities = [self.calculate_similarity(query_embedding, self.embedding_data[text]) for text in result_texts]
        most_relevant_idx = np.argmax(similarities)
        selected_results.append(self.embedding_data[result_texts[most_relevant_idx]])

        # Remove the most relevant result from the list of candidate results to avoid duplication
        result_texts.pop(most_relevant_idx)

        while len(selected_results) < top_k and len(result_texts) > 0:
            # Calculate the diversity score between each result and the selected results
            diversity_scores = [np.mean([self.calculate_similarity(emb_i, emb_j) for emb_i in selected_results]) for emb_j in [self.embedding_data[text] for text in result_texts]]

            # Calculate the diversified score (relevance - diversity * diversity_weight) for each result
            diversified_scores = [sim - diversity_weight * div_score for sim, div_score in zip(similarities, diversity_scores)]

            # Select the result with the highest diversified score
            next_idx = np.argmax(diversified_scores)

            # Add the selected result to the list of selected results
            selected_results.append(self.embedding_data[result_texts[next_idx]])

            # Remove the selected result from the list of candidate results to avoid duplication
            result_texts.pop(next_idx)

        return selected_results
    
    def reduce_embedding_dimension_pca(embeddings, target_dim):
        """
        Reduce the dimensionality of embeddings using Principal Component Analysis (PCA).

        Parameters:
            embeddings (numpy.ndarray): The embeddings matrix with shape (num_samples, original_dim).
            target_dim (int): The desired dimensionality of the reduced embeddings.

        Returns:
            numpy.ndarray: The reduced embeddings matrix with shape (num_samples, target_dim).
        """
        # Center the embeddings by subtracting the mean
        centered_embeddings = embeddings - np.mean(embeddings, axis=0)

        # Calculate the covariance matrix of the centered embeddings
        covariance_matrix = np.cov(centered_embeddings, rowvar=False)

        # Perform PCA to obtain the principal components and eigenvalues
        eigenvalues, eigenvectors = np.linalg.eigh(covariance_matrix)

        # Sort the eigenvalues and eigenvectors in descending order
        sorted_indices = np.argsort(eigenvalues)[::-1]
        sorted_eigenvalues = eigenvalues[sorted_indices]
        sorted_eigenvectors = eigenvectors[:, sorted_indices]

        # Select the top 'target_dim' eigenvectors (principal components) to reduce the dimensionality
        reduced_embeddings = np.dot(centered_embeddings, sorted_eigenvectors[:, :target_dim])

        # Add back the mean to obtain the final reduced embeddings
        final_reduced_embeddings = reduced_embeddings + np.mean(embeddings, axis=0)

        return final_reduced_embeddings
    
    def reduce_embedding_dimension_tsne(embeddings, target_dim):
        """
        Reduce the dimensionality of embeddings using t-SNE.

        Parameters:
            embeddings (numpy.ndarray): The embeddings matrix with shape (num_samples, original_dim).
            target_dim (int): The desired dimensionality of the reduced embeddings.

        Returns:
            numpy.ndarray: The reduced embeddings matrix with shape (num_samples, target_dim).
        """
        # Create a t-SNE model with the desired target dimension
        tsne_model = TSNE(n_components=target_dim)

        # Perform t-SNE to obtain the reduced embeddings
        reduced_embeddings = tsne_model.fit_transform(embeddings)

        return reduced_embeddings
    
    def reduce_embedding_dimension_umap(embeddings, target_dim):
        """
        Reduce the dimensionality of embeddings using UMAP.

        Parameters:
            embeddings (numpy.ndarray): The embeddings matrix with shape (num_samples, original_dim).
            target_dim (int): The desired dimensionality of the reduced embeddings.

        Returns:
            numpy.ndarray: The reduced embeddings matrix with shape (num_samples, target_dim).
        """
        # Create a UMAP model with the desired target dimension
        umap_model = umap.UMAP(n_components=target_dim)

        # Perform UMAP to obtain the reduced embeddings
        reduced_embeddings = umap_model.fit_transform(embeddings)

        return 
    
    def add_laplace_noise(self, epsilon):
        """
        Add Laplace noise to the embeddings to achieve differential privacy.

        Parameters:
            epsilon (float): The privacy parameter that controls the amount of noise added.

        Returns:
            None (The method modifies the 'embedding_data' dictionary in-place).
        """
        for key, embedding in self.embedding_data.items():
            noise = np.random.laplace(scale=1.0 / epsilon, size=embedding.shape)
            self.embedding_data[key] = embedding + noise

    def remove_laplace_noise(self):
        """
        Remove the previously added Laplace noise from the embeddings.

        Returns:
            None (The method modifies the 'embedding_data' dictionary in-place).
        """
        for key, embedding in self.embedding_data.items():
            self.embedding_data[key] = np.round(embedding)  # Remove the added noise by rounding the values


from typing import Any

from guardrail.guardchain.tools.base_tool import Tool

class HandOffToAgent(Tool):
    name: str = "Hand off" # Add type annotation 'str'
    description: str = "Hand off to a human agent"  # Add type annotation 'str'
    handoff_msg: str = "Let me hand you off to an agent now"  # Add type annotation 'str'

    def _run(self, *args: Any, **kwargs: Any) -> str:
        return self.handoff_msg



import logging
from typing import List, Optional

from transformers import (
    RobertaForSequenceClassification,
    RobertaTokenizer,
    TextClassificationPipeline,
)

from guardrail.firewall.util import device

from guardrail.firewall.output_detectors.base_detector import Detector

log = logging.getLogger(__name__)
model_name = "huggingface/CodeBERTa-language-id"


class LanguageDetection(Detector):
    """
    A detector to check for specific languages in text using Huggingface Transformers.
    """

    def __init__(self):
        """
        Initializes LanguageDetection with a language detection model.
        """
        self._pipeline = TextClassificationPipeline(
            model=RobertaForSequenceClassification.from_pretrained(model_name),
            tokenizer=RobertaTokenizer.from_pretrained(model_name),
            device=device,
        )
        log.debug(f"Initialized language detection model: {model_name}")

    def detect_languages(
        self, text: str, allowed: List[str], denied: List[str]
    ) -> (bool, float):
        """
        Checks whether the text contains languages listed in the 'allowed' or 'denied' lists.

        Parameters:
            text (str): The text to check.
            allowed (List[str]): A list of allowed languages.
            denied (List[str]): A list of denied languages.

        Returns:
            bool: True if the text contains allowed languages or doesn't contain denied languages. False otherwise.
            float: Risk score, where 0 is no risk and 1 is the highest risk.
        """
        if text.strip() == "":
            return True, 0.0

        languages = self._pipeline(text)
        log.debug(f"Detected languages: {languages}")

        for language in languages:
            language_name = language["label"]
            score = round(language["score"], 2)

            if len(allowed) > 0 and language_name in allowed:
                log.debug(f"Language {language_name} found in the allowed list with score {score}")
                return True, 0.0

            if len(denied) > 0 and language_name in denied:
                log.warning(f"Language {language_name} is not allowed (score {score})")
                return False, score

        if len(allowed) > 0:
            log.warning(f"No allowed languages detected: {languages}")
            return False, 1.0

        log.debug(f"No denied languages detected: {languages}")

        return True, 0.0


class CodingLanguageOutput(Detector):
    """
    A class for detecting if the output includes code in specific programming languages.
    """

    def __init__(self, allowed: Optional[List[str]] = None, denied: Optional[List[str]] = None):
        """
        Initializes CodingLanguage with the allowed and denied languages.

        Parameters:
            allowed (Optional[List[str]]): A list of allowed languages. Default is an empty list.
            denied (Optional[List[str]]): A list of denied languages. Default is an empty list.

        Raises:
            ValueError: If both 'allowed' and 'denied' lists are provided or if both are empty.
        """
        if not allowed:
            allowed = []

        if not denied:
            denied = []

        if len(allowed) > 0 and len(denied) > 0:
            raise ValueError("Provide either allowed or denied programming languages")

        if len(allowed) == 0 and len(denied) == 0:
            raise ValueError("No allowed or denied programming languages provided")

        self._allowed = allowed
        self._denied = denied

    def scan(self, prompt: str) -> (str, bool, float):
        language_detection = LanguageDetection()
        valid, score = language_detection.detect_languages(prompt, self._allowed, self._denied)
        return prompt, valid, score



import logging
from transformers import AutoModelForSequenceClassification, AutoTokenizer, TextClassificationPipeline
from guardrail.firewall.util import device
from guardrail.firewall.input_detectors.base_detector import Detector

_model_path = "JasperLS/gelectra-base-injection"
log = logging.getLogger(__name__)

class PromptInjections(Detector):

    def __init__(self, threshold: float = 0.75):
        model = AutoModelForSequenceClassification.from_pretrained(_model_path)
        self._tokenizer = AutoTokenizer.from_pretrained(_model_path)
        self._threshold = threshold
        self._text_classification_pipeline = TextClassificationPipeline(
            model=model,
            tokenizer=self._tokenizer,
            device=device,
        )
        log.debug(f"Initialized model {_model_path} on device {device}")

    def scan(self, prompt: str) -> (str, bool, float):
        if prompt.strip() == "":
            return prompt, True, 0.0

        result = self._text_classification_pipeline(
            prompt, truncation=True, max_length=self._tokenizer.model_max_length
        )
        injection_score = round(
            result[0]["score"] if result[0]["label"] == "INJECTION" else 1 - result[0]["score"], 2
        )

        if injection_score > self._threshold:
            log.warning(
                f"Detected prompt injection with score: {injection_score}, threshold: {self._threshold}"
            )
            return prompt, False, injection_score

        log.debug(
            f"No prompt injection detected (max score: {injection_score}, threshold: {self._threshold})"
        )
        return prompt, True, 0.0



import logging
from guardrail.firewall.input_detectors.base_detector import Detector
from textstat import textstat

log = logging.getLogger(__name__)

class TextQualityInput(Detector):
    """
    A detector to assess the quality of text using various text quality metrics.
    """

    def __init__(self):
        self.logger = logging.getLogger("TextQuality")
        logging.basicConfig(level=logging.INFO)

    def _scale_to_01(self, value, min_value, max_value):
        """
        Scale a value to the range [0, 1].

        Parameters:
            value (float): The value to be scaled.
            min_value (float): The minimum possible value.
            max_value (float): The maximum possible value.

        Returns:
            float: The scaled value in the range [0, 1].
        """
        return (value - min_value) / (max_value - min_value)

    def _evaluate(self, prompt):
        text = prompt
        metrics = {
            "automated_readability_index": {
                "col_type": "String",
                "schema_name": None,
                "min": 1.0,
                "max": 14.0,
                "function": lambda text: textstat.automated_readability_index(text),
            },
            "dale_chall_readability_score": {
                "col_type": "String",
                "schema_name": "text_standard_component",
                "min": 0.0,
                "max": 9.9,
                "function": lambda text: textstat.dale_chall_readability_score(text),
            },
            "linsear_write_formula": {
                "col_type": "String",
                "schema_name": "text_standard_component",
                "min": 0.0,
                "max": 12.0,
                "function": lambda text: textstat.linsear_write_formula(text),
            },
            "gunning_fog": {
                "col_type": "String",
                "schema_name": "text_standard_component",
                "min": 1.0,
                "max": 17.0,
                "function": lambda text: textstat.gunning_fog(text),
            },
            "aggregate_reading_level": {
                "col_type": "String",
                "schema_name": None,
                "function": lambda text: textstat.text_standard(text, float_output=True),
            },
            # Add more metrics here as needed
        }

        results = {}
        for metric, config in metrics.items():
            try:
                result = config["function"](text)
                self.logger.info(f"{metric}: {result}")
                # self.scale_to_01(result, min_value, max_value)  # Scale to [0, 1]
                results[metric] = result
            except Exception as e:
                self.logger.error(f"Error while evaluating {metric}: {e}")
                results[metric] = None

        return results

    def scan(self, prompt: str) -> (str, bool, float):
        # You can implement custom logic here to determine whether the text quality is acceptable or not
        # Example: If the average of scaled metrics is greater than 0.7, consider it high-quality text
        quality_metrics = self._evaluate(prompt)

        if quality_metrics["aggregate_reading_level"] > 20.0:
            is_high_quality = False
        elif quality_metrics["aggregate_reading_level"]:
            is_high_quality = True

        if is_high_quality:
            log.debug("Text is of high readability.")
            return prompt, True, quality_metrics
        else:
            log.warning("Text is of low readability.")
            return prompt, False, quality_metrics



import pdb
import requests
import asyncio
import aiohttp
import yaml
import os

class GoogleSerperAPIWrapper():
    """Wrapper around the Serper.dev Google Search API.
    You can create a free API key at https://serper.dev.
    To use, you should have the environment variable ``SERPER_API_KEY``
    set with your API key, or pass `serper_api_key` as a named parameter
    to the constructor.
    Example:
        .. code-block:: python
            from langchain import GoogleSerperAPIWrapper
            google_serper = GoogleSerperAPIWrapper()
    """
    def __init__(self, snippet_cnt = 10) -> None:
        self.k = snippet_cnt
        self.gl = "us"
        self.hl = "en"
        self.serper_api_key = os.environ.get("SERPER_API_KEY", None)
        assert self.serper_api_key is not None, "Please set the SERPER_API_KEY environment variable."
        assert self.serper_api_key != '', "Please set the SERPER_API_KEY environment variable."

    async def _google_serper_search_results(self, session, search_term: str, gl: str, hl: str) -> dict:
        headers = {
            "X-API-KEY": self.serper_api_key or "",
            "Content-Type": "application/json",
        }
        params = {"q": search_term, "gl": gl, "hl": hl}
        async with session.post(
            "https://google.serper.dev/search", headers=headers, params=params, raise_for_status=True
        ) as response:
            return await response.json()
    
    def _parse_results(self, results):
        snippets = []

        if results.get("answerBox"):
            answer_box = results.get("answerBox", {})
            if answer_box.get("answer"):
                element = {"content":answer_box.get("answer"),"source":"None"}
                return [element]
            elif answer_box.get("snippet"):
                element = {"content":answer_box.get("snippet").replace("\n", " "),"source":"None"}
                return [element]
            elif answer_box.get("snippetHighlighted"):
                element = {"content":answer_box.get("snippetHighlighted"),"source":"None"}
                return [element]
            
        if results.get("knowledgeGraph"):
            kg = results.get("knowledgeGraph", {})
            title = kg.get("title")
            entity_type = kg.get("type")
            if entity_type:
                element = {"content":f"{title}: {entity_type}","source":"None"}
                snippets.append(element)
            description = kg.get("description")
            if description:
                element = {"content":description,"source":"None"}
                snippets.append(element)
            for attribute, value in kg.get("attributes", {}).items():
                element = {"content":f"{attribute}: {value}","source":"None"}
                snippets.append(element)

        for result in results["organic"][: self.k]:
            if "snippet" in result:
                element = {"content":result["snippet"],"source":result["link"]}
                snippets.append(element)
            for attribute, value in result.get("attributes", {}).items():
                element = {"content":f"{attribute}: {value}","source":result["link"]}
                snippets.append(element)

        if len(snippets) == 0:
            element = {"content":"No good Google Search Result was found","source":"None"}
            return [element]
        
        # keep only the first k snippets
        snippets = snippets[:int(self.k / 2)]

        return snippets
    
    async def parallel_searches(self, search_queries, gl, hl):
        async with aiohttp.ClientSession() as session:
            tasks = [self._google_serper_search_results(session, query, gl, hl) for query in search_queries]
            search_results = await asyncio.gather(*tasks, return_exceptions=True)
            return search_results

    async def run(self, queries):
        """Run query through GoogleSearch and parse result."""
        flattened_queries = []

        for sublist in queries:
            if sublist is None:
                sublist = ['None', 'None']
            for item in sublist:
                flattened_queries.append(item)

        results = await self.parallel_searches(flattened_queries, gl=self.gl, hl=self.hl)
        snippets_list = []
        for i in range(len(results)):
            snippets_list.append(self._parse_results(results[i]))
        snippets_split = [snippets_list[i] + snippets_list[i+1] for i in range(0, len(snippets_list), 2)]
        return snippets_split

class google_search():
    def __init__(self, snippet_cnt):
        self.serper = GoogleSerperAPIWrapper(snippet_cnt=snippet_cnt)

    async def run(self, queries):
        print("Running google search...")
        print("Queries", queries)
        return await self.serper.run(queries)


from typing import List, Tuple

class Vault:
    """
    A utility class for storing tuples, often used as placeholder values for anonymization
    that need to be retained for later decoding.

    Provides methods for adding, removing, and retrieving tuples.

    Attributes:
        _tuples (List[Tuple]): The list of stored tuples.
    """

    def __init__(self):
        self._tuples: List[Tuple] = []

    def add(self, new_tuple: Tuple):
        """
        Adds a tuple to the vault.

        Args:
            new_tuple (Tuple): The tuple to be added.
        """
        self._tuples.append(new_tuple)

    def add_multiple(self, new_tuples: List[Tuple]):
        """
        Adds multiple tuples to the vault.

        Args:
            new_tuples (List[Tuple]): A list of tuples to be added.
        """
        self._tuples.extend(new_tuples)

    def remove(self, tuple_to_remove: Tuple):
        """
        Removes a specific tuple from the vault.

        Args:
            tuple_to_remove (Tuple): The tuple to be removed.
        """
        self._tuples.remove(tuple_to_remove)

    def get_all(self) -> List[Tuple]:
        """
        Retrieves all stored tuples from the vault.

        Returns:
            List[Tuple]: A list containing all the stored tuples.
        """
        return self._tuples
    
    def clear(self):
        """
        Resets the vault by clearing all stored tuples.
        """
        self._tuples.clear()



from typing import Protocol, Tuple

class Detector(Protocol):
    """
    Detector protocol defining the interface for detectors.

    Ensures detectors implement a common interface for interchangeable use.
    """

    def scan(self, output: str) -> Tuple[str, bool, float]:
        """
        Process and sanitize the output according to the specific detector's implementation.

        Args:
            out (str): The output to be processed.

        Returns:
            Tuple[str, bool, float]: A tuple containing:
                - str: The sanitized and processed output as per the detector's implementation.
                - bool: A flag indicating whether the output is valid or not.
                - float: Risk score where 0 means no risk and 1 means high risk.
        """



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Union

from pydantic import BaseModel

from guardrail.guardchain.agent.message import ChatMessageHistory, MessageType

class BaseMemory(BaseModel, ABC):
    """Base interface for memory in chains."""

    @abstractmethod
    def load_memory(
        self, key: Union[str, None] = None, default: Optional[Any] = None, **kwargs: Any
    ) -> Any:
        """Return key-value pairs given the text input to the chain."""

    @abstractmethod
    def load_conversation(self, **kwargs) -> ChatMessageHistory:
        """Return key-value pairs given the text input to the chain."""

    @abstractmethod
    def save_memory(self, key: str, value: Any) -> None:
        """Save the context of this model run to memory."""

    @abstractmethod
    def save_conversation(
        self, message: str, message_type: MessageType, **kwargs
    ) -> None:
        """Save the context of this model run to memory."""

    @abstractmethod
    def clear(self) -> None:
        """Clear memory contents."""



from langchain.evaluation.criteria.eval_chain import _SUPPORTED_CRITERIA

supported_criteria = {
    key: value.replace(" If so, response Y. If not, respond N.", '')
    if isinstance(value, str) else value
    for key, value in _SUPPORTED_CRITERIA.items()
}

LANGCHAIN_CORRECTNESS_PROMPT = f"{supported_criteria['correctness']} Respond only as a number from 1 to 10 where 1 is the least correct and 10 is the most correct."
LANGCHAIN_COHERENCE_PROMPT = f"{supported_criteria['coherence']} Respond only as a number from 1 to 10 where 1 is the least coherent and 10 is the most coherent."
LANGCHAIN_CONTROVERSIALITY_PROMPT = f"{supported_criteria['controversiality']} Respond only as a number from 1 to 10 where 1 is the most controversial and 10 is the least controversial."
LANGCHAIN_CONCISENESS_PROMPT = f"{supported_criteria['conciseness']} Respond only as a number from 1 to 10 where 1 is the least concise and 10 is the most concise."
LANGCHAIN_CRIMINALITY_PROMPT = f"{supported_criteria['criminality']} Respond only as a number from 1 to 10 where 1 is the most criminal and 10 is the least criminal."

LANGCHAIN_INSENSITIVITY_PROMPT = f"{supported_criteria['insensitivity']} Respond only as a number from 1 to 10 where 1 is the most insensitive and 10 is the least insensitive."
LANGCHAIN_HARMFULNESS_PROMPT = f"{supported_criteria['harmfulness']} Respond only as a number from 1 to 10 where 1 is the most harmful and 10 is the least harmful."
LANGCHAIN_HELPFULNESS_PROMPT = f"{supported_criteria['helpfulness']} Respond only as a number from 1 to 10 where 1 is the least helpful and 10 is the most helpful."

LANGCHAIN_MALICIOUSNESS_PROMPT = f"{supported_criteria['maliciousness']} Respond only as a number from 1 to 10 where 1 is the most malicious and 10 is the least malicious."
LANGCHAIN_MISOGYNY_PROMPT = f"{supported_criteria['misogyny']} If so, response Y. Respond only as a number from 1 to 10 where 1 is the most misogynistic and 10 is the least misogynistic."




import logging
import torch
from transformers import AutoModelForSequenceClassification, AutoTokenizer
from transformers import logging as transformers_logging

from guardrail.firewall.util import device

from .base_detector import Detector

_model_path = "ynie/roberta-large-snli_mnli_fever_anli_R1_R2_R3-nli"
transformers_logging.set_verbosity_error()
log = logging.getLogger(__name__)

MAX_LENGTH = 512

class Contradictions(Detector):
    """
    Contradictions Class:

    This class checks for contradictions between a given prompt and output using a pretrained NLI model.
    """

    def __init__(self, threshold=0.5):
        """
        Initializes an instance of the Contradictions class.

        Parameters:
            threshold (float): The threshold used to determine contradictions. Defaults to 0.
        """

        self._model = AutoModelForSequenceClassification.from_pretrained(_model_path)
        self._model.eval()
        self._model.to(device)
        self._tokenizer = AutoTokenizer.from_pretrained(_model_path)
        self._threshold = threshold

        log.debug(f"Initialized sentence transformer {_model_path} on device {device}")

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        if prompt.strip() == "":
            return output, True, 0.0

        tokenized_input_seq_pair = self._tokenizer.encode_plus(
            prompt,
            output,
            max_length=MAX_LENGTH,
            return_token_type_ids=True,
            truncation=True,
        )

        input_ids = (
            torch.Tensor(tokenized_input_seq_pair["input_ids"]).long().unsqueeze(0).to(device)
        )
        # Remember, some models like BART don't have 'token_type_ids'. Remove the line below if you are using such a model.
        token_type_ids = (
            torch.Tensor(tokenized_input_seq_pair["token_type_ids"]).long().unsqueeze(0).to(device)
        )
        attention_mask = (
            torch.Tensor(tokenized_input_seq_pair["attention_mask"]).long().unsqueeze(0).to(device)
        )

        outputs = self._model(
            input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids, labels=None
        )

        predicted_probability = torch.softmax(outputs[0], dim=1)[0].tolist()  # Batch size is only one

        log.debug(
            f"Detected entailment in the output with score: {predicted_probability[0]}, neutral: {predicted_probability[1]}, contradiction: {predicted_probability[2]}"
        )

        contradiction_score = round(predicted_probability[2], 2)
        if contradiction_score > self._threshold:
            log.warning(
                f"Detected contradiction in the output with score: {contradiction_score}, threshold: {self._threshold}"
            )

            return output, False, contradiction_score

        log.debug(
            f"No contradiction in the output. Max score: {contradiction_score}, threshold: {self._threshold}"
        )

        return output, True, 0.0



import logging
import tensorflow as tf
from genbit.genbit_metrics import GenBitMetrics
from transformers import AutoTokenizer, TFAutoModelForSequenceClassification
from guardrail.firewall.output_detectors.base_detector import Detector

_model_name = "d4data/bias-detection-model"
_language = "en"

log = logging.getLogger(__name__)

class Bias(Detector):
    """
    A detector to check for bias in text using GenBit and a Huggingface bias detection model.
    """

    def __init__(self, language_code=_language, model_name=_model_name):
        """
        Initializes BiasDetection with GenBit, a language code, and a Huggingface model for bias detection.

        Parameters:
            language_code (str): The language code for GenBit.
            model_name (str): The HuggingFace model name for bias detection.
        """
        self._genbit_metrics = None  # Initialize GenBitMetrics to None
        self._bias_detection_model = TFAutoModelForSequenceClassification.from_pretrained(model_name)
        self._tokenizer = AutoTokenizer.from_pretrained(model_name)

        log.debug(f"Initialized HuggingFace model: {model_name}")

    def scan(self, prompt: str, output: str) -> (str, bool, dict):
        """
        Detects bias in the given text.

        Parameters:
            text (str): The text to check for bias.

        Returns:
            tuple: A tuple containing:
                - str: The input text.
                - bool: True if bias is not detected in the text, False otherwise.
                - dict: A dictionary containing bias-related information (e.g., scores, metrics).
        """
        if output.strip() == "":
            return output, True, {}

        # Reinitialize GenBitMetrics for each scan to clear data
        self._genbit_metrics = GenBitMetrics(_language, context_window=5, distance_weight=0.95, percentile_cutoff=80)

        # Add text to GenBit for bias detection
        self._genbit_metrics.add_data([output], tokenized=False)

        # Generate gender bias metrics from GenBit
        metrics = self._genbit_metrics.get_metrics(output_statistics=True, output_word_list=True)
        log.debug(f"GenBit Bias metrics: {metrics}")

        # Tokenize and prepare input for the HuggingFace model
        # inputs = self._tokenizer(output, return_tensors="tf", padding=True, truncation=True)  # Use "tf" for TensorFlow
        # logits = self._bias_detection_model(inputs)[0]
        # bias_score = tf.nn.softmax(logits, axis=-1).numpy()[0]  # Assuming the model is multi-class

        # Check if bias is detected based on your criteria
        # You can customize this part based on your specific bias detection requirements
        bias_detected = False

        # Example: Check if bias is detected based on GenBit and HuggingFace model
        gender_metrics_score = metrics["genbit_score"]
        # Check your specific bias detection criteria here
        # Example: If the gender bias score is above a certain threshold and the model predicts bias, consider it as bias detected
        if gender_metrics_score > 0.5:
            bias_detected = True

        if bias_detected:
            log.warning("Bias detected in the output.")
        else:
            log.debug("No bias detected in the output.")

        genbit_metrics = {
            key: value for key, value in metrics.items() if key not in {"additional_metrics", "statistics", "token_based_metrics"}
        }

        result = {
            "bias_detected": bias_detected,
            "bias_score": gender_metrics_score,
            "genbit_metrics": genbit_metrics,
        }

        return output, not bias_detected, result



from transformers import AutoTokenizer, TFAutoModelForSequenceClassification
from transformers import pipeline


class Bias:
    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained("d4data/bias-detection-model")
        self.model = TFAutoModelForSequenceClassification.from_pretrained(
            "d4data/bias-detection-model"
        )

    def evaluate(self, text):
        classifier = pipeline(
            "text-classification", model=self.model, tokenizer=self.tokenizer
        )  # cuda = 0,1 based on gpu availability
        return classifier(text)



import pydantic 
from dotenv import load_dotenv
from pydantic_settings import BaseSettings

# Load environment variables from .env file
load_dotenv()

class GuardrailEnvConfig(BaseSettings, frozen=True):
    """Environment configurations for Guardrail"""

    guardrail_api_key: str = pydantic.Field(
        default="",
        env="GUARDRAIL_API_KEY",
        description="API key for Guardrail ML"
    )

    openai_api_key: str = pydantic.Field(
        default="",
        env="OPENAI_API_KEY",
        description="API key for OpenAI"
    )

    serper_api_key: str = pydantic.Field(
        default="",
        env="SERPER_API_KEY",
        description="API Key for Serper",
    )

    scraper_api_key: str = pydantic.Field(
        default="",
        env="SCRAPER_API_KEY",
        description="API Key for Scraper",
    )

    perspective_api_key: str = pydantic.Field(
        default="",
        env="PERSPECTIVE_API_KEY",
        description="API Key for Perspective",
    )

guardrail_env_config = GuardrailEnvConfig()


import logging
import re
from typing import List

from transformers import (
    AutoModelForSequenceClassification,
    AutoTokenizer,
    TextClassificationPipeline,
)

from guardrail.firewall.util import device

from .base_detector import Detector

_model_path = "elftsdmr/malware-url-detect"

log = logging.getLogger(__name__)

# Regular expression pattern to match URLs
url_pattern = re.compile(
    r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+"
)

class MalwareOutputURL(Detector):
    """
    MalwareURL Class:

    This class scans and detects malicious URLs in text using a pretrained model.
    It safeguards users from potential threats by classifying URLs as either malicious or benign.
    """

    def __init__(self, threshold=0.75):
        """
        Initializes an instance of the MalwareURL class.

        Parameters:
            threshold (float): The threshold used to determine if a URL is malicious. Defaults to 0.75.
        """

        model = AutoModelForSequenceClassification.from_pretrained(_model_path)
        self._tokenizer = AutoTokenizer.from_pretrained(_model_path)
        self._threshold = threshold
        self._text_classification_pipeline = TextClassificationPipeline(
            model=model,
            tokenizer=self._tokenizer,
            device=device,
        )
        log.debug(f"Initialized model {_model_path} on device {device}")

    @staticmethod
    def extract_urls(text: str) -> List[str]:
        """
        Extracts URLs from a given text.

        Args:
            text (str): The text to extract URLs from.

        Returns:
            List[str]: A list of URLs found in the text.
        """
        return url_pattern.findall(text)

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        if output.strip() == "":
            return output, True, 0.0

        urls = self.extract_urls(output)
        if len(urls) == 0:
            return output, True, 0.0

        log.debug(f"Found {len(urls)} URLs in the output")

        urls_str = ", ".join(urls)
        result = self._text_classification_pipeline(
            urls_str, truncation=True, padding=True, max_length=self._tokenizer.model_max_length
        )
        malware_score = (
            result[0]["score"] if result[0]["label"] == "MALWARE" else 1 - result[0]["score"]
        )
        if malware_score > self._threshold:
            log.warning(
                f"Detected a malware URL with score: {malware_score}, threshold: {self._threshold}"
            )

            return output, False, round(malware_score, 2)

        log.debug(
            f"No malware URLs detected in the output. Max score: {malware_score}, threshold: {self._threshold}"
        )

        return output, True, 0.0



def flatten_json(json_obj, parent_key='', separator='_'):
    flattened = {}
    for key, value in json_obj.items():
        new_key = parent_key + separator + key if parent_key else key
        if isinstance(value, dict):
            flattened.update(flatten_json(value, new_key, separator=separator))
        elif isinstance(value, list):
            if all(isinstance(item, dict) and 'label' in item and 'score' in item for item in value):
                for item in value:
                    flattened[new_key + '_label'] = str(item['label'])
                    flattened[new_key + '_score'] = str(item['score'])
            else:
                flattened[new_key] = str(value)
        else:
            flattened[new_key] = str(value)
    return flattened


import openai
import logging
from guardrail.metrics.utils.keys import load_api_key
from guardrail.firewall.input_detectors.base_detector import Detector

log = logging.getLogger(__name__)

class HarmfulInput(Detector):
    """
    A detector to detect harmful content using OpenAI's moderation API.
    """

    def __init__(self, thresholds=None):
        self._api_key = load_api_key("OPENAI_API_KEY")
        if thresholds == None:
            self.thresholds = {
                    "hate": 0.5,
                    "hate/threatening": 0.5,
                    "maliciousness": 0.5,
                    "controversiality": 0.5,
                    "coherence": 0.5,
                    "correctness": 0.5,
                    "harmfulness": 0.5,
                    "maliciousness": 0.5,
                    "helpfulness": 0.5,
                    "misogyny": 0.5,
                    "criminality": 0.5,
                    "insensitivity": 0.5
                }
        else: 
            self.thresholds = thresholds

    def _moderation(self, text: str) -> dict:
        openai.api_key = self._api_key
        return openai.Moderation.create(input=text)

    def scan(self, text: str) -> (str, bool, dict):
        if text.strip() == "":
            return text, True, {}

        moderation_result = self._moderation(text)

        if "results" in moderation_result:
            results = moderation_result["results"]
            if len(results) > 0:
                category_scores = results[0].get("category_scores", {})

                is_harmful = False
                scores = {}

                for category, threshold in self.thresholds.items():
                    score = category_scores.get(category, 0)
                    scores[category] = score
                    if score >= threshold:
                        is_harmful = True
                        log.warning(f"Content is harmful in category {category} (Score: {score})")

                if not is_harmful:
                    log.debug("Content is not harmful.")
                    return text, True, scores

        log.warning("Content is harmful.")
        return text, False, scores  # You can set the toxicity score to 1.0 for harmful content



import sqlite3


def insert_log(model_uri, prompt, output, metric_name, metric_value):
    # Create a database connection
    conn = sqlite3.connect("logs.db")

    # Insert the log into the database
    conn.execute(
        "INSERT INTO logs (timestamp, model_uri, prompt, output, metric_name, metric_value) VALUES (datetime('now'), ?, ?, ?, ?, ?)",
        (model_uri, prompt, output, metric_name, metric_value),
    )

    # Commit the changes and close the connection
    conn.commit()
    conn.close()



import torch
from transformers import AutoTokenizer, AutoModelForSequenceClassification


class Sentiment:
    def __init__(self, model_name="textattack/bert-base-uncased-imdb"):
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModelForSequenceClassification.from_pretrained(model_name)

    def evaluate(self, text, max_length=None):
        if max_length is None:
            # Use the default max_length of the model
            inputs = self.tokenizer(text, return_tensors="pt")
        else:
            # Set a custom max_length
            inputs = self.tokenizer(text, return_tensors="pt", max_length=max_length, truncation=True)

        outputs = self.model(**inputs)
        probabilities = torch.softmax(outputs.logits, dim=1)
        return probabilities[0, 1].item()



import time
from typing import Any, Dict, List

from langchain.chat_models import ChatOpenAI
from langchain.schema import LLMResult, HumanMessage
from langchain.callbacks.base import AsyncCallbackHandler, BaseCallbackHandler

from guardrail.tracker.openai import OpenAI

class GuardrailCallbackHandler(AsyncCallbackHandler):
    """Async callback handler that can be used to handle callbacks from langchain."""

    def __init__(self):
        super().__init__()
        self.prompts = []  # Initialize an empty list to store prompts
        self.openai_tracker = OpenAI()

    async def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> None:
        """Run when chain starts running."""
        print("Guardrail is tracking...")
        print(prompts)

        # Store prompts in the instance variable
        self.prompts = prompts

        # Record the start time
        self.start_time = time.time()

    async def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        """Run when chain ends running."""
        print("LLM chain is ending...")
        print(response)

        # Calculate the total time elapsed
        end_time = time.time()
        total_time = end_time - self.start_time
        print(f"Total time elapsed: {total_time} seconds")

        prompt = self.prompts[-1]
        chatbot_response = response.generations[0][0].text
        token_usage = response.llm_output["token_usage"]

        self.openai_tracker.run_eval_store_logs(prompt, chatbot_response, token_usage, total_time)



import streamlit as st
from enum import Enum
from helper.add_logo import add_logo
from PIL import Image
import pandas as pd
import numpy as np
from st_aggrid import AgGrid, GridOptionsBuilder
from helper.agstyler import PINLEFT, PRECISION_ONE, PRECISION_TWO, draw_grid, highlight

im = Image.open("favicon.ico")
st.set_page_config(page_title="Guardrail ML App", page_icon=im, layout="wide")

add_logo()

st.title("LLM Model Evaluation")
st.write("Average evaluation and task values displayed in the range from 0 (worst) to 1 (best).")

# Add custom CSS styles
st.markdown(
    """
    <style>
    .css-15ftl5c {
        color: white !important;
    }
    .css-t4mh3n {
        color: white !important;
    }
    table.dataframe {
        width: 100%;
        height: 100%;
        padding-top: -20px;
        border-collapse: collapse;
    }
    div[data-testid="stTable"] table {
        width: 100%;
        height: 100%;
    }
    table.dataframe th, table.dataframe td {
        padding: 0px;
        text-align: center;
    }
    table.dataframe th {
        background-color: #555555;
        color: white;
        font-weight: bold;
    }
    table.dataframe td {
        background-color: white;
        font-weight: normal;
        border: none;
    }
    .table-content {
        margin: 10px;
    }
    </style>
    """,
    unsafe_allow_html=True,
)

# st.sidebar.success("Select a page above.")

models = [
    "dolly-v2-12b",
    "alpaca-lora-7b",
    "open_llama_13b",
    "mosaic-mpt-7b",
    "vicuna-13b",
    "gpt4",
]

# Create a sample dataframe
data = {
    "Base Model": np.random.choice(models, size=30),
    "Version": np.random.uniform(1.0, 3.0, size=30),
    "Relevance": np.random.uniform(0, 1, size=30),
    "Toxicity": np.random.uniform(0, 1, size=30),
    "Bias": np.random.uniform(0, 1, size=30),
    "Quality": np.random.uniform(0, 1, size=30),
    "Security": np.random.uniform(0, 1, size=30),
    "Sentiment": np.random.uniform(0, 1, size=30),
    "TruthfulQA (MC) (0-s)": np.random.uniform(0, 1, size=30),
}


class Color(Enum):
    RED_LIGHT = "#fcccbb"
    GREEN_LIGHT = "#abf7b1"


condition_one_value = "params.value < 0.5"
condition_other_values = "params.value > 0.8"

formatter = {
    "Base Model": ("Base Model", {"width": 80}),
    "Version": ("Version", {**PRECISION_ONE, "width": 40}),
    "Relevance": (
        "Relevance",
        {
            **PRECISION_TWO,
            "width": 50,
            "cellStyle": highlight(Color.GREEN_LIGHT.value, condition_other_values),
        },
    ),
    "Toxicity": (
        "Toxicity",
        {
            **PRECISION_TWO,
            "width": 50,
            "cellStyle": highlight(Color.GREEN_LIGHT.value, condition_other_values),
        },
    ),
    "Bias": (
        "Bias",
        {
            **PRECISION_TWO,
            "width": 50,
            "cellStyle": highlight(Color.RED_LIGHT.value, condition_one_value),
        },
    ),
    "Quality": (
        "Quality",
        {
            **PRECISION_TWO,
            "width": 50,
            "cellStyle": highlight(Color.GREEN_LIGHT.value, condition_other_values),
        },
    ),
    "Security": (
        "Security",
        {
            **PRECISION_TWO,
            "width": 50,
            "cellStyle": highlight(Color.RED_LIGHT.value, condition_one_value),
        },
    ),
    "Sentiment": (
        "Sentiment",
        {
            **PRECISION_TWO,
            "width": 50,
            "cellStyle": highlight(Color.GREEN_LIGHT.value, condition_other_values),
        },
    ),
    "TruthfulQA (MC) (0-s)": (
        "TruthfulQA (MC) (0-s)",
        {
            **PRECISION_TWO,
            "width": 50,
            "cellStyle": highlight(Color.GREEN_LIGHT.value, condition_other_values),
        },
    ),
}

# row_number = st.number_input('Number of rows', min_value=0, value=30)

df = pd.DataFrame(data)
data = draw_grid(
    df,
    formatter=formatter,
    fit_columns=True,
    selection="multiple",  # or 'single', or None
    use_checkbox="True",  # or False by default
    max_height=3000,
)



import logging
import concurrent.futures
import spacy.cli
from typing import Dict, List, Tuple

from guardrail.metrics.utils.keys import init_keys, init_guardrail_key
from guardrail.firewall.input_detectors.base_detector import Detector as InputDetector
from guardrail.firewall.output_detectors.base_detector import Detector as OutputDetector

from guardrail.firewall.vault import Vault
from guardrail.firewall.input_detectors import Anonymize, Secrets, PromptInjections, DoSTokens, StopInputSubstrings, MalwareInputURL, HarmfulInput, TextQualityInput
from guardrail.firewall.output_detectors import Deanonymize, SensitivePII, StopOutputSubstrings, FactualityDetector, Contradictions, MalwareOutputURL, HarmfulOutput, TextQualityOutput, Bias, Relevance

logging.basicConfig(level=logging.INFO)
model_name = "en_core_web_trf"

class Firewall():
    def __init__(self, foundation_model="gpt-3.5-turbo", no_defaults=False):
        init_keys()
        self.foundation_model = foundation_model
        self.vault = Vault()
        if not no_defaults:
            spacy.cli.download(model_name)
            self.default_input_detectors =  [Anonymize(self.vault), Secrets(), PromptInjections(), DoSTokens(), StopInputSubstrings(), MalwareInputURL(), HarmfulInput(), TextQualityInput()]
            self.default_output_detectors = [SensitivePII(), Deanonymize(self.vault), StopOutputSubstrings(), Contradictions(), MalwareOutputURL(), HarmfulOutput(), TextQualityOutput(), Bias(), Relevance()]

    def clear_input_detectors(self):
        """Clear the list of input detectors."""
        self.default_input_detectors.clear()

    def clear_output_detectors(self):
        """Clear the list of output detectors."""
        self.default_output_detectors.clear()

    def add_to_input_detectors(self, detector: InputDetector):
        """
        Add an input detector to the list of input detectors.

        Args:
            detector (InputDetector): An instance of an input detector class.
        """
        self.default_input_detectors.append(detector)

    def add_to_output_detectors(self, detector: OutputDetector):
        """
        Add an output detector to the list of output detectors.

        Args:
            detector (OutputDetector): An instance of an output detector class.
        """
        self.default_output_detectors.append(detector)
    
    def print_detectors(self):
        print("Input Detectors: ", self.default_input_detectors)
        print("Output Detectors: ", self.default_output_detectors)

    def scan_input(self,
                   prompt: str,
                   detectors: List[InputDetector] = None,
                   ) -> Tuple[str, Dict[str, bool], Dict[str, float]]:
        """
        Scans a given prompt using the provided detectors or default detectors.

        Args:
            detectors (List[InputDetector], optional): A list of detector objects. Each detector should be an instance of a class that inherits from `Detector`. Defaults to None.
            prompt (str, optional): The input prompt string to be scanned. Defaults to an empty string.

        Returns:
            Tuple[str, Dict[str, bool], Dict[str, float]]: A tuple containing:
                - The processed prompt string after applying all detectors.
                - A dictionary mapping detector names to boolean values indicating whether the input prompt is valid according to each detector.
                - A dictionary mapping detector names to float values of risk scores, where 0 is no risk, and 1 is high risk.
        """

        sanitized_prompt = prompt
        results_valid = {}
        results_score = {}

        if detectors is None:
            detectors = self.default_input_detectors

        if len(detectors) == 0 or prompt.strip() == "":
            return sanitized_prompt, results_valid, results_score

        # Create a ThreadPoolExecutor to parallelize detector execution
        with concurrent.futures.ThreadPoolExecutor() as executor:
            # Submit detector jobs and collect futures
            futures = {executor.submit(detector.scan, sanitized_prompt): detector for detector in detectors}

            # Collect results from completed futures
            for future in concurrent.futures.as_completed(futures):
                detector = futures[future]
                sanitized_prompt, is_valid, risk_score = future.result()
                results_valid[type(detector).__name__] = is_valid
                results_score[type(detector).__name__] = risk_score

        return sanitized_prompt, results_valid, results_score

    def scan_output(self,
                    prompt: str,
                    output: str,
                    detectors: List[OutputDetector] = None,
                    ) -> Tuple[str, Dict[str, bool], Dict[str, float]]:
        """
        Scans a given output of a large language model using the provided detectors or default detectors.

        Args:
            detectors (List[OutputDectector], optional): A list of detector objects. Each detector should be an instance of a class that inherits from `Detector`. Defaults to None.
            prompt (str, optional): The input prompt string that produced the output. Defaults to an empty string.
            output (str, optional): The output string to be scanned. Defaults to an empty string.

        Returns:
            Tuple[str, Dict[str, bool], Dict[str, float]]: A tuple containing:
                - The processed output string after applying all detectors.
                - A dictionary mapping detector names to boolean values indicating whether the output is valid according to each detector.
                - A dictionary mapping detector names to float values of risk scores, where 0 is no risk, and 1 is high risk.
        """

        sanitized_output = output
        results_valid: Dict[str, bool] = {}
        results_score: Dict[str, float] = {}

        if detectors is None:
            detectors = self.default_output_detectors

        if not detectors or not output.strip():
            return sanitized_output, results_valid, results_score

        # Create a ThreadPoolExecutor to parallelize detector execution
        with concurrent.futures.ThreadPoolExecutor() as executor:
            # Submit detector jobs and collect futures
            futures = {executor.submit(detector.scan, prompt, sanitized_output): detector for detector in detectors}

            # Collect results from completed futures
            for future in concurrent.futures.as_completed(futures):
                detector = futures[future]
                sanitized_output, is_valid, risk_score = future.result()
                results_valid[type(detector).__name__] = is_valid
                results_score[type(detector).__name__] = risk_score

        return sanitized_output, results_valid, results_score



import base64

import pkg_resources
import streamlit as st


def add_logo():
    logo = open(pkg_resources.resource_filename("helper", "guardrail_logo.svg"), "rb").read()
    logo_encoded = base64.b64encode(logo).decode()

    st.markdown(
        f"""
        <style>
            [data-testid="stSidebarNav"] {{
                background-image: url('data:image/svg+xml;base64,{logo_encoded}');
                background-size: 100% auto;
                padding-top: 60px;
                background-position: 0px 0px;
                background-repeat: no-repeat;
            }}
            [data-testid="grSidebarNav"]::before {{
                margin-left: 0px;
                margin-top: 0px;
                font-size: 30px;
                position: relative;
                top: 100px;
            }}
        </style>
        """,
        unsafe_allow_html=True,
    )



import logging
from guardrail.metrics.offline_evals import Textstat, Toxicity, Relevance, PromptInjection, Sentiment, Bias

class OfflineMetricsEvaluator:
    def __init__(self):
        # Initialize the metrics classes
        self.textstat = Textstat()
        self.toxicity = Toxicity()
        self.relevance = Relevance()
        self.injection = PromptInjection()
        self.sentiment = Sentiment()
        self.bias = Bias()
        self.logger = logging.getLogger("Textstat")
        print("Initialized complete...")

    def evaluate_metrics(self, prompt, output):
        results = {}

        ts_results = self.textstat.evaluate(output)
        sentiment_results = self.sentiment.evaluate(output, max_length=1024)
        bias_results = self.bias.evaluate(output)
        
        if prompt:
            relevance_results = self.relevance.evaluate(output, prompt)
            injection_results = self.injection.evaluate(prompt)
            toxicity_results = self.toxicity.evaluate(output, prompt)

        results["text_quality"] = ts_results
        results["toxicity"] = toxicity_results
        results["sentiment"] = sentiment_results
        results["bias"] = bias_results

        if relevance_results:
            results["relevance"] = relevance_results
        if injection_results:
            results["prompt_injection"] = injection_results

        return results



from typing import Any, Optional, Dict

from pydantic import BaseModel, Field

from guardrail.guardchain.agent.message import ChatMessageHistory, MessageType
from guardrail.guardchain.memory.base import BaseMemory

class BufferMemory(BaseMemory, BaseModel):
    """Buffer for storing conversation memory and an in-memory kv store."""

    conversation_history: ChatMessageHistory = Field(default_factory=ChatMessageHistory)
    kv_memory: Dict[str, Any] = Field(default_factory=dict)

    def load_memory(
        self, key: Optional[str] = None, default: Optional[Any] = None, **kwargs
    ) -> Any:
        """Return history buffer by key or all memories."""
        if not key:
            return self.kv_memory

        return self.kv_memory.get(key, default)

    def load_conversation(self, **kwargs) -> ChatMessageHistory:
        """Return history buffer and format it into a conversational string format."""
        return self.conversation_history

    def save_memory(self, key: str, value: Any) -> None:
        self.kv_memory[key] = value

    def save_conversation(
        self, message: str, message_type: MessageType, **kwargs
    ) -> None:
        """Save context from this conversation to buffer."""
        self.conversation_history.save_message(
            message=message, message_type=message_type, **kwargs
        )

    def clear(self) -> None:
        """Clear memory contents."""
        self.conversation_history.clear()

    class Config:
        # Set the `allow_extra` attribute to True to disable validation of extra fields
        allow_extra = True



from transformers import AutoTokenizer, AutoModelForSequenceClassification
from scipy.spatial.distance import cosine


class Relevance:
    def __init__(self, model_name="sentence-transformers/paraphrase-MiniLM-L6-v2"):
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModelForSequenceClassification.from_pretrained(model_name)

    def evaluate(self, text1, text2):
        inputs1 = self.tokenizer(text1, return_tensors="pt")
        inputs2 = self.tokenizer(text2, return_tensors="pt")
        outputs1 = self.model(**inputs1)
        outputs2 = self.model(**inputs2)
        similarity = 1 - cosine(
            outputs1.logits.flatten().detach().numpy(),
            outputs2.logits.flatten().detach().numpy(),
        )
        return similarity



import torch
from transformers import AutoTokenizer, AutoModelForSequenceClassification


class Toxicity:
    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained("nicholasKluge/ToxicityModel")
        self.model = AutoModelForSequenceClassification.from_pretrained(
            "nicholasKluge/ToxicityModel"
        )
        self.model.eval()

    def evaluate(self, response, prompt):
        inputs = self.tokenizer(
            prompt,
            response,
            truncation=True,
            return_token_type_ids=False,
            return_attention_mask=True,
            return_tensors="pt",
        )
        outputs = self.model(**inputs)
        score = self.model(**inputs)[0]
        # probabilities = torch.softmax(score.logits, dim=0)
        return score.item()



from __future__ import annotations

import logging
from string import Template
from typing import Any, Dict, List, Optional, Union

from colorama import Fore

from guardrail.guardchain.agent.base_agent import BaseAgent
from guardrail.guardchain.agent.prompts import (
    CLARIFYING_QUESTION_PROMPT_TEMPLATE,
    PLANNING_PROMPT_TEMPLATE,
    SHOULD_ANSWER_PROMPT_TEMPLATE,
    SAFE_PROMPT_TEMPLATE,
    FIX_TOOL_INPUT_PROMPT_TEMPLATE,
)
from guardrail.guardchain.agent.message import BaseMessage, ChatMessageHistory, UserMessage
from guardrail.guardchain.agent.base_agent import ConvoJSONOutputParser, JSONPromptTemplate
from guardrail.guardchain.agent.agent_structs import AgentAction, AgentFinish
from guardrail.guardchain.models.base_model import BaseLanguageModel, Generation
from guardrail.guardchain.tools.base_tool import Tool
from guardrail.guardchain.utils import print_with_color

logger = logging.getLogger(__name__)


class ChatAgent(BaseAgent):
    output_parser: ConvoJSONOutputParser = ConvoJSONOutputParser()
    llm: Any = None
    prompt_template: JSONPromptTemplate = None
    allowed_tools: Dict[str, Tool] = {}
    tools: List[Tool] = []
    prompt: str = ""

    @classmethod
    def from_llm_and_tools(
        cls,
        llm: Any,
        tools: Optional[List[Tool]] = None,
        output_parser: Optional[ConvoJSONOutputParser] = None,
        prompt_template: str = PLANNING_PROMPT_TEMPLATE,
        input_variables: Optional[List[str]] = None,
        prompt: str = "",
        **kwargs: Any,
    ) -> ChatAgent:
        tools = tools or []
        template = cls.get_prompt_template(
            template=prompt_template,
            input_variables=input_variables,
        )
        allowed_tools = {tool.name: tool for tool in tools}
        _output_parser = output_parser or ConvoJSONOutputParser()
        return cls(
            llm=llm,
            allowed_tools=allowed_tools,
            output_parser=_output_parser,
            prompt_template=template,
            tools=tools,
            prompt=prompt,
            **kwargs,
        )

    def should_answer(
        self,
        should_answer_prompt_template: str = SAFE_PROMPT_TEMPLATE,
        **kwargs,
    ) -> Optional[AgentFinish]:
        if "history" not in kwargs or not kwargs["history"]:
            return None
        history = kwargs.pop("history")
        inputs = {
            "history": history.format_message(),
            **kwargs,
        }

        def _parse_response(res: dict):
            if 'safe_prompt' in res and res['safe_prompt'] == 'no':
                return AgentFinish(
                    message="Thank you for contacting",
                    log="Thank you for contacting",
                )
            else:
                return None

        prompt = Template(should_answer_prompt_template).substitute(**inputs)
        response = self.llm.generate([UserMessage(content=prompt)]).generations[0].message.content
        return _parse_response(response)

    @staticmethod
    def format_prompt(
        template: JSONPromptTemplate,
        intermediate_steps: List[AgentAction],
        **kwargs: Any,
    ) -> List[BaseMessage]:
        def _construct_scratchpad(actions: List[AgentAction]) -> Union[str, List[BaseMessage]]:
            scratchpad = ""
            for action in actions:
                scratchpad += action.response
            return scratchpad

        thoughts = _construct_scratchpad(intermediate_steps)
        new_inputs = {"agent_scratchpad": thoughts}
        full_inputs = {**kwargs, **new_inputs}
        prompt = template.format_prompt(**full_inputs)
        return prompt

    @staticmethod
    def get_prompt_template(
        template: str = "",
        input_variables: Optional[List[str]] = None,
    ) -> JSONPromptTemplate:
        template = Template(template)
        if input_variables is None:
            input_variables = ["input", "agent_scratchpad"]
        return JSONPromptTemplate(template=template, input_variables=input_variables)

    def plan(
        self,
        history: ChatMessageHistory,
        intermediate_steps: List[AgentAction],
        **kwargs: Any,
    ) -> Union[AgentAction, AgentFinish]:
        print_with_color("Planning the next steps...", Fore.LIGHTYELLOW_EX)
        tool_names = ", ".join([tool.name for tool in self.tools])
        tool_strings = "\n\n".join([f"> {tool.name}: \n{tool.description}" for tool in self.tools])
        inputs = {
            "tool_names": tool_names,
            "tools": tool_strings,
            "history": history.format_message(),
            "prompt": self.prompt,
            "human_input": history.get_latest_user_message().content,
            **kwargs,
        }
        final_prompt = self.format_prompt(self.prompt_template, intermediate_steps, **inputs)
        print("Final prompt: ", final_prompt)
        print_with_color("finish with tooling", Fore.CYAN)
        logger.info(f"\nPlanning Input: {final_prompt[0].content} \n")
        full_output: Generation = self.llm.generate(final_prompt, planning=True).generations[0]
        print("Full output", full_output)
        agent_output: Union[AgentAction, AgentFinish] = self.output_parser.parse(
            full_output.message, tool_names, llm=self.llm
        )
        print_with_color("done with tooling", Fore.CYAN)

        print(f"Planning output: \n{repr(full_output.message.content)}", Fore.YELLOW)
        if isinstance(agent_output, AgentAction):
            print_with_color(
                f"Plan to take the following action: '{agent_output.tool}'", Fore.LIGHTYELLOW_EX
            )

        return agent_output

    def clarify_args_for_agent_action(
        self,
        agent_action: AgentAction,
        history: ChatMessageHistory,
        intermediate_steps: List[AgentAction],
        **kwargs: Any,
    ):
        print_with_color("Deciding if I need additional clarification...", Fore.LIGHTYELLOW_EX)
        if not self.allowed_tools.get(agent_action.tool):
            return agent_action
        else:
            inputs = {
                "tool_name": agent_action.tool,
                "tool_desp": self.allowed_tools.get(agent_action.tool).description,
                "history": history.format_message(),
                **kwargs,
            }
            clarifying_template = self.get_prompt_template(
                template=CLARIFYING_QUESTION_PROMPT_TEMPLATE
            )
            final_prompt = self.format_prompt(clarifying_template, intermediate_steps, **inputs)
            print("Final prompt: ", final_prompt)
            logger.info(f"\nClarification inputs: {final_prompt[0].content}")
            full_output: Generation = self.llm.generate(final_prompt).generations[0]
            print(f"Clarification outputs: {repr(full_output.message.content)}")
            return self.output_parser.parse_clarification(
                full_output.message, agent_action=agent_action, llm=self.llm
            )

    def fix_action_input(self, tool: Tool, action: AgentAction, error: str) -> AgentAction:
        prompt = FIX_TOOL_INPUT_PROMPT_TEMPLATE.format(
            tool_description=tool.description, inputs=action.tool_input, error=error
        )
        logger.info(f"\nFixing tool input prompt: {prompt}")
        messages = UserMessage(content=prompt)
        output = self.llm.generate([messages]).generations[0]
        new_tool_inputs = self.output_parser.load_json_output(output.message, self.llm)

        logger.info(f"\nFixed tool output: {new_tool_inputs}")
        new_action = AgentAction(tool=action.tool, tool_input=new_tool_inputs)
        return new_action



"""
1. run_metrics
2. create_dataset 
3. init_logs 
"""

import logging
import sqlite3
from dotenv import set_key

from .metrics.offline_evals import Bias, Toxicity, Relevance, PromptInjection, Sentiment, Textstat

from .db import insert_log

def set_env_vars(file_path='.env', variables_dict={}):
    """
    Set environment variables programmatically in a .env file.

    Args:
        file_path (str): The path to the .env file.
        variables_dict (dict): A dictionary where keys are variable names and values are their corresponding values.
    """
    for variable, value in variables_dict.items():
        set_key(file_path, variable, value)

def run_simple_metrics(output, prompt, model_uri):
    # Initialize the Textstat class
    textstat = Textstat()

    # Configure logging
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger("Textstat")

    ts_eval_results = textstat.evaluate(output)

    for result_name in ts_eval_results:
        try:
            # Insert log into the database
            insert_log(model_uri, prompt, output, result_name, ts_eval_results[result_name])
        except Exception as e:
            logger.error(f"Error while inserting {result_name} into DB: {e}")

    return ts_eval_results



def run_metrics(output, prompt, model_uri):
    # Initialize the metrics classes
    textstat = Textstat()
    toxicity = Toxicity()
    relevance = Relevance()
    injection = PromptInjection()
    sentiment = Sentiment()
    bias = Bias()

    # Configure logging
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger("Textstat")

    results = {}
    ts_results = textstat.evaluate(output)
    sentiment_results = sentiment.evaluate(output)
    bias_results = bias.evaluate(output)

    if prompt:
        relevance_results = relevance.evaluate(output, prompt)
        injection_results = injection.evaluate(prompt)
        toxicity_results = toxicity.evaluate(output, prompt)

    results["text_quality"] = ts_results
    for result_name in ts_results:
        try:
            # Insert log into the database
            insert_log(
                model_uri,
                prompt,
                output,
                "tq_" + str(result_name),
                ts_results[result_name],
            )
        except Exception as e:
            logger.error(f"Error while inserting {result_name} into DB: {e}")

    insert_log(model_uri, prompt, output, "toxicity", toxicity_results)
    insert_log(model_uri, prompt, output, "sentiment", sentiment_results)
    insert_log(model_uri, prompt, output, "bias_label", bias_results[0]["label"])
    insert_log(model_uri, prompt, output, "bias_score", bias_results[0]["score"])

    results["toxicity"] = toxicity_results
    results["sentiment"] = sentiment_results
    results["bias"] = bias_results

    if relevance_results:
        results["relevance"] = relevance_results
        insert_log(model_uri, prompt, output, "relevance", relevance_results)
    if injection_results:
        results["prompt_injection"] = injection_results
        insert_log(model_uri, prompt, output, "prompt_injection", injection_results)
    return results





import logging
import json
from typing import Optional

from googleapiclient import discovery
from guardrail.metrics.utils.keys import load_api_key, init_perspective_key

from guardrail.firewall.input_detectors.base_detector import Detector

log = logging.getLogger(__name__)

class ToxicityInput(Detector):
    """
    A detector to detect and prevent toxic comments using the Perspective API.
    """

    def __init__(self):
        """
        Initializes the ToxicityDetector with a Perspective API key.

        Parameters:
            api_key (str): Your Perspective API key.
        """
        self._api_key = load_api_key("PERSPECTIVE_API_KEY")
        init_perspective_key()

        # Create a client for the Perspective API
        self._client = discovery.build(
            "commentanalyzer",
            "v1alpha1",
            developerKey=self._api_key,
            discoveryServiceUrl="https://commentanalyzer.googleapis.com/$discovery/rest?version=v1alpha1",
            static_discovery=False,
        )

    def _analyze_toxicity(self, text: str) -> float:
        """
        Analyzes the toxicity of a given text using the Perspective API.

        Parameters:
            text (str): The text to analyze for toxicity.

        Returns:
            float: The toxicity score (0.0 to 1.0) indicating the level of toxicity.
        """
        analyze_request = {
            'comment': {'text': text},
            'requestedAttributes': {'TOXICITY': {}}
        }

        response = self._client.comments().analyze(body=analyze_request).execute()
        toxicity_score = response.get('attributeScores', {}).get('TOXICITY', {}).get('summaryScore', {}).get('value', 0.0)
        return toxicity_score

    def scan(self, comment_text: str) -> (str, bool, float):
        if comment_text.strip() == "":
            return comment_text, True, 0.0

        toxicity_score = self._analyze_toxicity(comment_text)

        if toxicity_score < 0.5:
            log.debug(f"Comment is not toxic (Toxicity Score: {toxicity_score})")
            return comment_text, True, toxicity_score

        log.warning(f"Comment is toxic (Toxicity Score: {toxicity_score})")
        return comment_text, False, toxicity_score



from __future__ import annotations

import logging
import json
from abc import ABC, abstractmethod
from typing import Any, List, Optional, Sequence, Union
from string import Template

from pydantic import BaseModel, Extra, Field
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from transformers import AutoTokenizer, AutoModelForCausalLM

from guardrail.guardchain.agent.message import BaseMessage, UserMessage
from guardrail.guardchain.agent.agent_structs import AgentAction, AgentFinish, AgentOutputParser
from guardrail.guardchain.models.base_model import BaseLanguageModel
from guardrail.guardchain.tools.base_tool import Tool


logger = logging.getLogger(__name__)


class JSONPromptTemplate(BaseModel):
    """Format prompt with string Template and dictionary of variables."""

    template: Template
    """The prompt template."""
    
    class Config:
        """Configuration for this pydantic object."""
        extra = Extra.forbid
        arbitrary_types_allowed = True
        validate_assignment = False  # Turn off Pydantic validation errors

    input_variables: List[str]
    """A list of the names of the variables the prompt template expects."""

    def format_prompt(self, **kwargs: Any) -> List[BaseMessage]:
        variables = {v: "" for v in self.input_variables}
        variables.update(kwargs)
        prompt = self.template.substitute(**variables)
        return [UserMessage(content=prompt)]


class BaseAgent(BaseModel, ABC):
    output_parser: AgentOutputParser = None
    llm: Any = None
    tools: Sequence[Tool] = []

    class Config:
        """Configuration for this pydantic object."""
        extra = Extra.forbid
        arbitrary_types_allowed = True
        validate_assignment = False  # Turn off Pydantic validation errors

    @classmethod
    def from_llm_and_tools(
        cls,
        llm: Any,
        tools: Sequence[Tool],
        prompt: str,
        output_parser: Optional[AgentOutputParser] = None,
        input_variables: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> BaseAgent:
        """Construct an agent from an LLM and tools."""
        pass

    def should_answer(
        self, should_answer_prompt_template: str = "", **kwargs
    ) -> Optional[AgentFinish]:
        """Determine if agent should continue to answer user questions based on the latest user query."""
        return None
    
    def safe_prompt(
        self, safe_prompt_template: str = "", **kwargs
    ) -> Optional[AgentFinish]:
        """Determine if agent should continue to answer user questions based on the safety of the input prompt."""
        return None

    @abstractmethod
    def plan(
        self,
        history: ChatMessageHistory,
        intermediate_steps: List[AgentAction],
        **kwargs: Any,
    ) -> Union[AgentAction, AgentFinish]:
        """
        Plan the next step, either taking an action with AgentAction or respond to the user with AgentFinish.
        Args:
            history: Entire conversation history between user and agent including the latest query.
            intermediate_steps: List of AgentAction that has been performed with outputs.
            **kwargs: Key-value pairs from chain, which contains query and other stored memories.

        Returns:
            AgentAction or AgentFinish
        """
        pass

    def clarify_args_for_agent_action(
        self,
        agent_action: AgentAction,
        history: ChatMessageHistory,
        intermediate_steps: List[AgentAction],
        **kwargs: Any,
    ) -> Union[AgentAction, AgentFinish]:
        """
        Ask clarifying question if needed. When the agent is about to perform an action, we could
        use this function with a different prompt to ask clarifying questions for input if needed.
        Sometimes the planning response would already have the clarifying question, but we found
        it is more precise if there is a different prompt just for clarifying args.

        Args:
            agent_action: Agent action about to take.
            history: Conversation history including the latest query.
            intermediate_steps: List of agent actions taken so far.
            **kwargs:

        Returns:
            Either a clarifying question (AgentFinish) or take the planned action (AgentAction)
        """
        pass

    def fix_action_input(
        self, tool: Tool, action: AgentAction, error: str
    ) -> Optional[AgentAction]:
        """If the tool failed due to an error, what should be the fix for inputs."""
        pass

    @staticmethod
    def get_prompt_template(
        prompt: str = "",
        input_variables: Optional[List[str]] = None,
    ) -> JSONPromptTemplate:
        """
        Create a prompt in the style of the zero-shot agent.

        Args:
            prompt: Message to be injected between prefix and suffix.
            input_variables: List of input variables the final prompt will expect.

        Returns:
            A PromptTemplate with the template assembled from the pieces here.
        """
        pass



class ConvoJSONOutputParser(AgentOutputParser):
    def is_json(self, data):
        try:
            json.loads(data)
            return True
        except json.JSONDecodeError:
            return False

    def ensure_json(self, data):
        if isinstance(data, dict):
            # If data is already a dictionary (JSON object), return as it is
            return data
        elif self.is_json(data):
            # If data is a valid JSON string, parse and return as a dictionary
            return json.loads(data)
        else:
            raise ValueError("Data is not in valid JSON format.")
        
    def parse(self, message: BaseMessage, tool_names, llm) -> Union[AgentAction, AgentFinish]:
        if self.is_json(message.content):
            response = message.content
        else:
            response = self.load_json_output(message, llm)
            if not self.is_json(response):
                response = self.ensure_json(response)

        action_name = response.get("tool", {}).get("name")
        action_args = response.get("tool", {}).get("args")

        if (
            not response.get("thoughts", {}).get("need_use_tool")
            or not action_name
            or action_name not in tool_names
        ):
            output_message = response.get("assistant").get("response")
            if output_message:
                return AgentFinish(message=output_message, log=output_message)
            else:
                return AgentFinish(message="Sorry, I don't understand", log=output_message)

        return AgentAction(
            tool=action_name,
            tool_input=action_args,
            model_response=response.get("response", ""),
        )

    def parse_clarification(
        self, message: BaseMessage, agent_action: AgentAction, llm
    ) -> Union[AgentAction, AgentFinish]:
        response = self.load_json_output(message, llm)

        has_arg_value = response.get("has_arg_value", "")
        clarifying_question = response.get("clarifying_question", "")

        if "no" in has_arg_value.lower() and clarifying_question:
            return AgentFinish(message=clarifying_question, log=clarifying_question)
        else:
            return agent_action



import os
import openai
from dotenv import load_dotenv, find_dotenv
from guardrail.client import set_env_vars
import os

class InvalidApiKeyError(Exception):
    pass

def load_api_key(env_var_name):
    """
    Load an API key from an environment variable.

    Parameters:
        env_var_name (str): The name of the environment variable containing the API key.

    Returns:
        str: The API key.
    """
    api_key = os.getenv(env_var_name)
    if not api_key:
        raise ValueError(f"{env_var_name} not found in environment variables.")
    return api_key

def init_openai_key():
    """
    Initialize the OpenAI API key from the environment variable.
    """
    openai_api_key = load_api_key("OPENAI_API_KEY")
    openai.api_key = openai_api_key

def init_guardrail_key(key=""):
    """
    Initialize the Guardrail API key from the environment variable.
    """
    if key == "":
        guardrail_api_key = load_api_key("GUARDRAIL_API_KEY")
        if guardrail_api_key == "":
            raise InvalidApiKeyError("Please enter a valid Guardrail API Key.")   
    else:
        guardrail_api_key = key
    set_env_vars({"GUARDRAIL_API_KEY": guardrail_api_key})

def init_perspective_key():
    """
    Initialize the Perspective API key from the environment variable.
    """
    perspective_api_key = load_api_key("PERSPECTIVE_API_KEY")

    set_env_vars({"PERSPECTIVE_API_KEY": perspective_api_key})

def init_keys(api_keys=None):
    """
    Initialize API keys for OpenAI and Guardrail.

    Parameters:
        api_keys (dict, optional): A dictionary containing API keys. If provided, it will set the environment variables.
    """
    if api_keys:
        set_env_vars(api_keys)
    init_guardrail_key()
    init_openai_key()


import PyPDF2
from transformers import AutoTokenizer, LlamaTokenizer
import json
import requests
import re

# 1. Change Tokenizer
# 2. Change instruction_template
# 3. Change PDF file

HOST = "localhost:5000"
URI = f"http://{HOST}/api/v1/chat"
tokenizer = LlamaTokenizer.from_pretrained("psmathur/orca_mini_v2_7b")
history = {"internal": [], "visible": []}
command = (
    "You are an API that converts bodies of text into a single question and answer into a JSON format. Each JSON "
    "contains a single question with a single answer that is referenceable. Only respond with the JSON and no additional text. \n"
)


def run(user_input, history):
    request = {
        "user_input": user_input,
        "history": history,
        "mode": "instruct",  # Valid options: 'chat', 'chat-instruct', 'instruct'
        "character": "Example",
        "instruction_template": "Orca Mini",
        "your_name": "Kevin",
        "regenerate": False,
        "_continue": False,
        "stop_at_newline": False,
        "chat_prompt_size": 2048,
        "chat_generation_attempts": 1,
        "chat-instruct_command": "",
        "max_new_tokens": 500,
        "do_sample": True,
        "temperature": 0.7,
        "top_p": 0.1,
        "typical_p": 1,
        "epsilon_cutoff": 0,  # In units of 1e-4
        "eta_cutoff": 0,  # In units of 1e-4
        "tfs": 1,
        "top_a": 0,
        "repetition_penalty": 1.18,
        "top_k": 40,
        "min_length": 0,
        "no_repeat_ngram_size": 0,
        "num_beams": 1,
        "penalty_alpha": 0,
        "length_penalty": 1,
        "early_stopping": False,
        "mirostat_mode": 0,
        "mirostat_tau": 5,
        "mirostat_eta": 0.1,
        "seed": -1,
        "add_bos_token": True,
        "truncation_length": 500,
        "ban_eos_token": False,
        "skip_special_tokens": True,
        "stopping_strings": [],
    }

    response = requests.post(URI, json=request)

    result = response.json()["results"][0]["history"]
    return result["visible"][-1][1]


def extract_text_from_pdf(file_path):
    pdf_file_obj = open(file_path, "rb")
    pdf_reader = PyPDF2.PdfReader(pdf_file_obj)
    text = ""
    for page_num in range(len(pdf_reader.pages)):
        page_obj = pdf_reader.pages[page_num]
        text += page_obj.extract_text()
    pdf_file_obj.close()
    return text


def tokenize(text):
    enc = tokenizer.encode(text)
    return enc


def chunks(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def is_json(data):
    try:
        json.loads(data)
        return True
    except ValueError:
        return False


def submit_to_api(chunk, retries=3):
    for i in range(retries):
        try:
            response = run(command + chunk.strip(), history)
            # Extract JSON string from between back-ticks
            print("initial response", response)
            if is_json(response):
                print(response)
                return json.loads(response)
            else:
                match = re.search(r"`(.*?)`", response, re.S)
                if match and is_json(match.group(1)):
                    print(f"Attempt {i + 1} failed. Retrying...")
                    return json.loads(match.group(1))  # assuming you want to return the JSON data
        except requests.exceptions.RequestException as e:
            print(f"Request failed: {e}")
            continue
    print("Max retries exceeded. Skipping this chunk.")
    return None


def tokenize_pdf(path):
    text = extract_text_from_pdf(path)
    tokens = tokenize(text)

    token_chunks = list(chunks(tokens, 256))
    print("Number of token_chunks", len(token_chunks))
    return token_chunks


def get_responses(token_chunks, responses):
    for chunk in token_chunks:
        response = submit_to_api(tokenizer.decode(chunk))
        if response is not None:
            responses.append(response)

    # Write responses to a JSON file
    with open("responses.json", "w") as f:
        json.dump(responses, f)


token_chunks = tokenize_pdf("./constitution.pdf")
responses = get_responses(token_chunks, [])
